<?php

/**
 * @file
 * Functions and classes for the all entities.
 */

define('LIMESURVEY_SYNC_SYNCHRONIZED', 1);
define('LIMESURVEY_SYNC_DESYNCHRONIZED', 2);
define('LIMESURVEY_SYNC_UNSYNCHRONIZED', 3);

// Fieldname for LS question type '1' contain an '#" that break Drupal queries.
// I replace it.
define('LIMESURVEY_SYNC_REPLACE_HASK_KEY', 'z');

/**
 * Class managing building dynamic entities.
 *
 * Building entity_info, entity_property_info, schema, entity default values and
 * updating entity table if needed. Take care of Drupal fields, LimeSurvey fixed
 * fields and Limesurver variables fields.
 */
class LimesurveySyncDpInfosLs extends LimesurveySyncDpInfos {
  // 'survey', or 'response' or 'answer' or 'token'.
  protected $ls_entity_type;
  protected $module = 'limesurvey_sync';
  // If the entity contains 'lss_sync' synchronized field.
  protected $ls_sync_col = TRUE;
  protected $ls_status_col = TRUE;
  protected $promote_col = TRUE;
  // The LS primary key.
  protected $ls_key = '';
  // The LS storing date.
  protected $ls_date_key = '';

  // Uri for the overview page.
  public $root_admin_ui = 'admin/structure/limesurvey';
  // Uri for the view page.
  public $crud_path = 'limesurvey';

  // If the entity contains 'language' language field.
  protected $language_col = FALSE;
  // If the entity manage revision.
  protected $revision = TRUE;
  // If the entity is fieldable.
  protected $fieldable = TRUE;
  // If the entity is exportable.
  protected $exportable = FALSE; // --> add manually 'rules controller class' in entity_info.
  // Correspond to entity_info['bundle of'].

  /**
   * Set the new object.
   */
  public function __construct() {
    if (empty($this->entityname) || $this->entityname == 'entityname') {
      $this->entityname = limesurvey_sync_api_entityname($this->ls_entity_type);
    }
    if (empty($this->tablename) || empty($this->tablename_revision)) {
      $this->tablename = limesurvey_sync_api_tablename($this->ls_entity_type);
      $this->tablename_revision = limesurvey_sync_api_tablename($this->ls_entity_type, FALSE, TRUE);
    }
    if (empty($this->crud_path)) {
      $this->crud_path = limesurvey_sync_api_typename($this->ls_entity_type);
    }
    $for_classes = drupal_ucfirst(str_replace('_', '', $this->ls_entity_type));
    if (empty($this->entityclassname)) {
      $this->entityclassname = "LimesurveySyncEntityLs{$for_classes}";
    }
    if (empty($this->controller_class)) {
      $this->controller_class = "LimesurveySync{$for_classes}Controller";
    }
    if (empty($this->ui_controller_class)) {
      $this->ui_controller_class = "LimesurveySync{$for_classes}UIController";
    }
    parent::__construct();
  }

  /** Inform the system about one or more entity types
   *
   * Building entity_info corresponding to settings above.
   *
   *   @see hook_entity_info()
   */
  public function entity_info() {

    $entities_info = parent::entity_info();

    if ($this->ls_key) {
      // Add custom entries : corresponding LS primary key.
      $entities_info[$this->entityname]['entity keys']['ls key'] = $this->ls_key;
    }
    if ($this->ls_date_key) {
      // Add custom entries : LS primary last change date field.
      $entities_info[$this->entityname]['entity keys']['ls date key'] = $this->ls_date_key;
    }

    return $entities_info;
  }

  /**
   * Define schema for custom fields.
   *
   * each class should implement it.
   *
   * @return
   *   An array, the custom schema fields or an empty array if there is
   *   no fields.
   *
   * @see LimesurveySyncDpInfos::schema()
   */
  protected function schema_custom_fields($revision = NULL, $datas = array()) {
    $fields = parent::schema_custom_fields($revision);
    if ($this->ls_sync_col) {
      // Add custom entries : corresponding LS primary key.
      $fields['lss_sync'] = array(
        'type' => 'int',
        'unsigned' => TRUE,
        'not NULL' => TRUE,
        'default' => 1,
        'size' => 'tiny',
        'description' => "Sync status of this {$this->label} (1 for synchronized, 2 for desynchronized, 3 for unsynchronized).",
      );
    }

    return array_merge($fields, $this->schema_ls_fields($datas));
  }

  /**
   * Define schema for LS fields.
   *
   * Merge known and variable LS fields schema, returning all LS fields schema.
   * If a LS field exists on LS but is not defined on Drupal, provide a watchdog
   * information about it.
   *
   * @param $sid
   *   A numeric, the LS survey ID.
   *
   * @return
   *   An array, the exhaustive LS schema fields or an empty array if there is
   *   no LS fields.
   *
   * @see LimesurveySyncInfos::schema()
   */
  protected function schema_ls_fields($datas = array()) {

    // Set here for performance raisons. Thoses tables don't need update.
    if (in_array($this->ls_entity_type, array('survey', 'response'))) {
      return array();
    }
    $sid = (!empty($datas['ls_sid'])) ? $datas['ls_sid'] : NULL;
    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
    $existing_ls_fields = limesurvey_sync_ls_api_retrieve_ls_fields($this->ls_entity_type, $sid);
    // If the token table does not exists on limesurvey, return nothing.
    $ls_current_fields = $existing_ls_fields;
    $fields_schema = array();
    foreach ($ls_current_fields as $key => $current_ls_field) {
      $field_schema = $this->schema_ls_fields_known($current_ls_field);
      if ($field_schema) {
        $fields_schema[$current_ls_field] = $field_schema;
        unset($existing_ls_fields[$key]);
      }
    }

    if (!empty($existing_ls_fields)) {
      $remaining_fields = $existing_ls_fields;
      foreach ($remaining_fields as $key => $variable_ls_field) {

        $field_schema = $this->schema_ls_fields_variables($variable_ls_field, $sid);
        if ($field_schema) {
          $fields_schema[$variable_ls_field] = $field_schema;
          unset($existing_ls_fields[$key]);
        }
      }
    }

    if (!empty($existing_ls_fields) && !db_table_exists(limesurvey_sync_api_tablename($this->ls_entity_type, $sid))) {
      $arg = array(
        '%fields' => implode(', ', $existing_ls_fields),
        '%sid' => $sid,
        '%type' => $this->ls_entity_type,
      );
      watchdog('limesurvey_sync', 'the LimeSurvey fields %fields have not been imported into Drupal (from %type %sid).', $arg, WATCHDOG_WARNING);
      drupal_set_message(t('the LimeSurvey fields %fields have not been imported into Drupal (from %type %sid).', $arg), 'warning');
    }

    return $fields_schema;
  }

  /**
   * Define schema for a fixed (known) LS field.
   *
   * For token : Handled tid, completed, firstname, etc... fields and for answer
   * handled id, lastpage, startlanguage, etc.. fields.
   *
   * @param $field
   *   A string, the LS field name.
   *
   * @return
   *   An array, the schema field or FALSE if it is not a fixed field.
   *
   * @see LimesurveySyncInfos::schema()
   */
  protected function schema_ls_fields_known($field) {
    return FALSE;
  }

  /**
   * Return the corresponding Drupal database schema of a LimeSurvey field.
   *
   * This fonction handles LimeSurvey token or answer fields that are specific to
   * a particular LimeSurvey survey. For the token table it corresponds to the
   * attribute fields, and for the answer table it corresponds to the questions
   * fields. The LimeSurvey datetime fields will be converted to timestamp on
   * Drupal.
   *
   * @param $field
   *   A string, the LS field name.
   * @param $sid
   *   A numeric, the LS survey ID.
   *
   * @return
   *   An array of Drupal schema corresponding to the provided LimeSurvey field or
   *   FALSE if the provided field is not a variable LimeSurvey field.
   *
   * @see LimesurveySyncInfos::schema()
   */
  protected function schema_ls_fields_variables($field, $sid) {
    return FALSE;
  }

  /**
   * Provided entity_property_info() informations for custom fields.
   *
   * Each class should implement it.
   *
   * @return
   *   An array, the DP entity_property_info() fields.
   *
   * @see LimesurveySyncDpInfos::entity_property_info()
   */
  protected function entity_property_info_custom_fields($datas = array()) {

    // No html for label.
    // set some prefix for sorting :
    // '_ID' for dp and LS ids.
    // 'DP'for drupal fields.
    // 'LS admin for ls admin fields.
    // 'LS datas' then the field name without prefix for answer datas and token
    // attributes.
    // No need to add the survey ID : provided by views.
    $sid = (!empty($datas['ls_sid'])) ? $datas['ls_sid'] : NULL;
    $survey_arg = (!empty($sid)) ? array(
      '@survey' => t(' (survey #@sid)', array(
        '@sid' => $sid,
      )),
    ) : array();
    $arg_label = array_merge(array(
      '@type' => $this->ls_entity_type,
    ), $survey_arg);
    $survey_desc = (!empty($sid)) ? array(
      '%survey' => t(' (survey #@sid)', array(
        '@sid' => $sid,
      )),
    ) : array();
    $arg_desc = array_merge(array(
      '%type' => $this->ls_entity_type,
    ), $survey_desc);


    $fields = parent::entity_property_info_custom_fields($datas);
    // Add the ls_entity_type_field.
    $fields['ls_entity_type'] = array(
      'label' => t('LimeSurvey entity type'),
      'description' => t('The type of LimeSurvey entity (%type)', $arg_desc),
      'type' => 'text',
      'entity views field' => TRUE,
    );
    if ($this->ls_sync_col) {
      $fields['lss_sync'] = array(
        'label' => t('DP synchronization status'),
        'description' => t('sync ID (1 for synchronized, 2 for desynchronized, 3 for unsynchronized)'),
        'type' => 'integer',
        'schema field' => 'lss_sync',
        'options list' => 'limesurvey_sync_api_lss_sync_status_list',
        'setter permission' => "administer limesurvey sync {$this->ls_entity_type}",
      );
    }
    $fields = array_merge($fields, $this->entity_property_info_ls_fields($sid));

    return $fields;
  }

  /**
   * Provided entity_property_info() informations for LS known fields.
   *
   * For token : Handled tid, completed, firstname, etc... fields and for answer
   * handled id, lastpage, startlanguage, etc.. fields.
   *
   * @param $field
   *   A string, the LS field name.
   * @param $sid
   *   A numeric, the LS survey ID.
   *
   * @return
   *   An array, the entity_property_info() field or FALSE if it is not a fixed
   *   field.
   *
   * @see LimesurveySyncInfos::entity_property_info()
   */
  protected function entity_property_info_ls_fields_known($field, $sid) {
    return FALSE;
  }

  /**
   * Provided entity_property_info() informations for LS variable fields.
   *
   * For token : Handled attribute_1, attribute_2, etc... fields and for answer
   * handled 79447X7X1731, 79447X7X1732, etc.. fields.
   *
   * @param $field
   *   A string, the LS field name.
   * @param $sid
   *   A numeric, the LS survey ID.
   *
   * @return
   *   An array, the entity_property_info() field or FALSE if it is not an
   *   existing variable field.
   *
   * @see LimesurveySyncInfos::entity_property_info()
   */
  protected function entity_property_info_ls_fields_variables($field, $sid) {
    return FALSE;
  }

  /**
   * Provided entity_property_info() informations for LS fields.
   *
   * Merge known and variable LS fields entity_property_info, returning all LS
   * fields entity_property_info. If a LS field exists on LS but is not defined
   * on Drupal, provide a watchdog information about it.
   *
   * @param $sid
   *   A numeric, the LS survey ID. Default is NULL.
   *
   * @return
   *   An array, the exhaustive LS entity_property_info() fields or an empty
   *   array if there is no LS fields.
   *
   * @see LimesurveySyncInfos::entity_property_info()
   */
  protected function entity_property_info_ls_fields($sid = NULL) {

    if (!in_array($this->ls_entity_type, array(
      'token',
      'answer',
    ))) {
      return array();
    }

    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
    $existing_ls_fields = limesurvey_sync_ls_api_retrieve_ls_fields($this->ls_entity_type, $sid);
    $ls_current_fields = $existing_ls_fields;
    $fields_entity_property_info = array();
    foreach ($ls_current_fields as $key => $current_ls_field) {
      $field_entity_property_info = $this->entity_property_info_ls_fields_known($current_ls_field, $sid);
      if ($field_entity_property_info) {
        $fields_entity_property_info[$current_ls_field] = $field_entity_property_info;
        unset($existing_ls_fields[$key]);
      }
    }

    if (!empty($existing_ls_fields)) {
      $remaining_fields = $existing_ls_fields;
      foreach ($remaining_fields as $key => $variable_ls_field) {
        $field_entity_property_info = $this->entity_property_info_ls_fields_variables($variable_ls_field, $sid);
        if ($field_entity_property_info) {
          $fields_entity_property_info[$variable_ls_field] = $field_entity_property_info;
          unset($existing_ls_fields[$key]);
        }
      }
    }

    if (!empty($existing_ls_fields) && !db_table_exists(limesurvey_sync_api_tablename($this->ls_entity_type, $sid))) {
      $arg = array(
        '%fields' => implode(', ', $existing_ls_fields),
        '%sid' => $sid,
        '%type' => $this->ls_entity_type,
      );
      watchdog('limesurvey_sync', 'the LimeSurvey fields %fields have no defined properties into Drupal (from %type %sid).', $arg, WATCHDOG_WARNING);
      drupal_set_message(t('the LimeSurvey fields %fields have no defined properties into Drupal (from %type %sid).', $arg), 'warning');
    }

    // Add ls_sid field.
    $ls_sid_field = array(
      'ls_sid' => array(
        'label' => t('_ID, LimeSurvey survey ID'),
        'description' => t('LimeSurvey survey ID into the LimeSurvey database.'),
        'type' => 'integer',
        'validation callback' => 'entity_metadata_validate_integer_positive',
        'entity views field' => TRUE,
      ),
    );
    $fields_entity_property_info += $ls_sid_field;
    return (!empty($fields_entity_property_info)) ? $fields_entity_property_info : array();
  }


  /**
   * Provide default values on entity custom fields.
   *
   * Each class should implement it.
   *
   * @return
   *   An array, the defaut values keyed by there field name.
   */
  protected function entity_default_values_custom($datas = array()) {
    $default_values = parent::entity_default_values_custom();
    $default_values['ls_sync'] = LIMESURVEY_SYNC_UNSYNCHRONIZED;
    return $default_values;
  }

  /**
   * Provide default values on creating entity.
   *
   * @param $sid
   *   A numeric, the LS survey ID. Default is NULL.
   *
   * @return
   *   An array, the defaut values keyed by there field name.
   *
   * @see LimesurveySyncInfos::entity_default_values()
   */
  protected function create_dp_fields_default_values($ls_sid = NULL) {

    return array();
  }

  /**
   * Provide default values on creating entity for LS fields.
   *
   * Merge known and variable LS fields default values. If a LS field exists on
   * LS but is not defined on Drupal, provide a watchdog information about it.
   *
   * @param $sid
   *   A numeric, the LS survey ID. Default is NULL.
   *
   * @return
   *   An array, the exhaustive defaut values keyed by there field name for LS
   *   fields.
   *
   * @see LimesurveySyncInfos::entity_default_values()
   */
  protected function create_ls_fields_default_values($sid = NULL) {

    if (!in_array($this->ls_entity_type, array(
      'token',
      'answer',
    ))) {
      return array();
    }

    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
    $existing_ls_fields = limesurvey_sync_ls_api_retrieve_ls_fields($this->ls_entity_type, $sid);
    $ls_current_fields = $existing_ls_fields;
    $default_values = array();
    foreach ($ls_current_fields as $key => $current_ls_field) {
      $field_default_values = $this->create_ls_fields_default_values_known($current_ls_field, $sid);
      if ($field_default_values !== FALSE) {
        $default_values[$current_ls_field] = $field_default_values;
        unset($existing_ls_fields[$key]);
      }
    }

    if (!empty($existing_ls_fields)) {
      $remaining_fields = $existing_ls_fields;
      foreach ($remaining_fields as $key => $variable_ls_field) {
        $field_default_values = $this->create_ls_fields_default_values_variables($variable_ls_field, $sid);
        if ($field_default_values !== FALSE) {
          $default_values[$variable_ls_field] = $field_default_values;
          unset($existing_ls_fields[$key]);
        }
      }
    }

    if (!empty($existing_ls_fields) && !db_table_exists(limesurvey_sync_api_tablename($this->ls_entity_type, $sid))) {
      $arg = array(
        '%fields' => implode(', ', $existing_ls_fields),
        '%sid' => $sid,
        '%type' => $this->ls_entity_type,
      );
      watchdog('limesurvey_sync', 'the LimeSurvey fields %fields have no default value into Drupal (from %type %sid).', $arg, WATCHDOG_WARNING);
      drupal_set_message(t('the LimeSurvey fields %fields have no default value into Drupal (from %type %sid).', $arg), 'warning');
    }

    return (!empty($default_values)) ? $default_values : array();
  }

  /**
   * Return the corresponding Drupal default value of a LS known field.
   *
   * This fonction is used for creating a new answer or token Drupal entity.
   * Handle common LimeSurvey fields of the token and answer tables.
   *
   * @param $field
   *   A string. The LimeSurvey field name.
   * @param $sid
   *   A numeric. The LimeSurvey survey ID. Default is NULL.
   *
   * @return
   *   A string or numeric or NULL : the Drupal default value for the provided
   *   LimeSurvey field or FALSE if the provided field is not a generic (known)
   *   LimeSurvey field.
   *
   * @see LimesurveySyncInfos::entity_default_values()
   */
  protected function create_ls_fields_default_values_known($field, $sid = NULL) {
    return FALSE;
  }

  /**
   * Return the corresponding Drupal default value of a LS variable field.
   *
   * This fonction is used for creating a new answer or token Drupal entity.
   * Handle variable LimeSurvey fields of the token and answer tables.
   *
   * @param $field
   *   A string. The LimeSurvey field name.
   * @param $sid
   *   A numeric. The LimeSurvey survey ID. Default is NULL.
   *
   * @return
   *   A string or numeric or NULL : the Drupal default value for the provided
   *   LimeSurvey field or FALSE if the provided field is not a variable
   *   LimeSurvey field.
   *
   * @see LimesurveySyncInfos::entity_default_values()
   */
  protected function create_ls_fields_default_values_variables($field, $sid = NULL) {
    return FALSE;
  }

  /**
   * Update schema fields regarding changes on the LS database.
   *
   * This method is called on saving a survey entity, compare the current schema
   * stored into a variable and add colomns if needed (as an example : when a
   * question has been added to a survey on the LS admin interface). Set a
   * watchdog entry if add colomns. For the response entity : add corresponding
   * entity reference fields.
   *
   * @return
   *   A boolean, TRUE if the schema has been updated (so the cached schema need
   *   to be rebuild).
   *
   * @see LimesurveySyncInfos::schema()
   * @see LimesurveySyncSurveyController::save()
   */
  public function update_table($datas = array()) {

    $variable_name = "{$this->tablename}_ls_fields";
    $stored_ls_fields = variable_get($variable_name, array());
    $ls_fields = $this->schema_ls_fields($datas);
    $ls_fields_list = array_keys($ls_fields);
    $diff = array_diff($ls_fields_list, $stored_ls_fields);
    $update = $rebuild_schema = FALSE;
    // It is important to test the base table first, then the revision table.
    $iterate = array($this->tablename);
    if ($this->revision) {
      $iterate[] = $this->tablename_revision;
    }
    // Schema() make changes on $this !
    $schema = (!empty($datas['schema'])) ? $datas['schema'] : $this->schema($datas);
    foreach ($iterate as $tablename) {
      $info_schema = $schema[$tablename];
      if (!db_table_exists($tablename)) {
        // Table does not exists : save it.
        db_create_table($tablename, $info_schema);
        // save it into a variable.
        $update = TRUE;
        $rebuild_schema = TRUE;
        watchdog('limesurvey_sync', 'the LimeSurvey sync table %table has been created.', array('%table' => $tablename), WATCHDOG_NOTICE);
      }
      else {
        // The table exists : look for changes on fields.
        if (!empty($diff)) {
          // Somes colomns have been added from LS.
          $added_field = array();
          foreach ($diff as $field) {
            if (!db_field_exists($tablename, $field)) {
              db_add_field($tablename, $field, $info_schema['fields'][$field]);
              $added_field[] = $field;
            }
          }
          // save it into a variable.
          $update = TRUE;
          if (count($added_field) > 0) {
            $rebuild_schema = TRUE;
            watchdog('limesurvey_sync', 'the LimeSurvey sync table %table has been updated, fields added : %fields.', array(
              '%table' => $tablename,
              '%fields' => implode(', ', $added_field),
            ), WATCHDOG_NOTICE);
          }
        }
        // Check for deleted fields on LS :
        $diff_del = array_diff($stored_ls_fields, $ls_fields_list);
        if (!empty($diff)) {
          // Somes colomns have been deleted from LS.
          $deleted_field = array();
          foreach ($diff as $field) {
            if (!db_field_exists($tablename, $field)) {
              // Delete this colomn only if there all values are all NULL into DP.
              $query =  db_select($tablename, 't');
              $query->addField('t', $field);
              $result = $query->execute();
              $field_values = $result->fetchCol();
              $values = array_values(array_unique($field_values));
              if (count($values) < 1 || (count($values) == 1 && array_key_exists('default', $info_schema['fields'][$field]) &&  $values[0] == $info_schema['fields'][$field]['default'])) {
                db_drop_field($tablename, $field);
                $deleted_field[] = $field;
              }
            }
          }
          // save it into a variable.
          $update = TRUE;
          if (count($deleted_field) > 0) {
            $rebuild_schema = TRUE;
            watchdog('limesurvey_sync', 'the LimeSurvey sync table %table has been updated, fields deleted : %fields.', array(
              '%table' => $tablename,
              '%fields' => implode(', ', $deleted_field),
            ), WATCHDOG_NOTICE);
          }
        }

      }
    }

    // Tables have been created or updated.
    if ($update) {
      // Create or update the bundle.
      // Save the new schema into a variable. Thanks to this, if a new
      // question or attribute is added from LimeSurvey, the system will be
      // able to detect it and add it into Drupal.
      // Save only the base table schema.
      $stored_ls_fields = variable_set($variable_name, $ls_fields_list);

    }

    if ($rebuild_schema) {
      drupal_get_schema(NULL, TRUE);
    }
    // Return TRUE if the schema as changed.
    return $rebuild_schema;
  }

}

class LimesurveySyncDpInfosLsSubentity extends LimesurveySyncDpInfosLs {
  /**
   * Set the new object : answers and tokens.
   */
  // Array of drupal ls_sids keyed by type.
  protected $ls_sids = array();
  protected $ls_sid = FALSE;
  protected $ls_sids_valid = array();
  protected $entity_label_callback = 'limesurvey_sync_api_sub_label_callback';

  public function __construct() {
    // No user , no title_col col.
    $this->user_col = FALSE;
    $this->title_col = FALSE;
    // If the entity can be published/unpublished.
    $this->status_col = FALSE;
    // If the comment are open or not.
    $this->comment_col = FALSE;
    // If the entity can be promoted to front page.
    $this->promote_col = FALSE;
    // If the entity can be set as sticky.
    $this->sticky_col = FALSE;

    // custom config
    $this->fieldable = FALSE;
    //$this->exportable = TRUE;
    $this->perm_add = TRUE; // FALSE
    $this->perm_update = TRUE; // FALSE
    $this->perm_view = TRUE;
    $this->perm_delete = TRUE;
    $this->revision = TRUE;

    // Retrieve surveys.
    // Bypass entity_load() as we cannot use it during entity_info build.
    $this->ls_sids = array();
    $survey_table = limesurvey_sync_api_tablename('survey');
    if (db_table_exists($survey_table)) {
      $query = db_query('SELECT ls_sid, type FROM {' . $survey_table . '} ORDER by ls_sid');
      while ($row = $query->fetchAssoc()) {
        $this->ls_sids[$row['type']] = $row['ls_sid'];
      }
    }
    parent::__construct();
  }

  protected function set_for_ls_sid() {
    if (!empty($this->ls_sid)) {
      $this->entityname = limesurvey_sync_api_entityname($this->ls_entity_type, $this->ls_sid);
      // Rewrite tables names.
      $this->tablename = limesurvey_sync_api_tablename($this->ls_entity_type, $this->ls_sid);
      $this->tablename_revision = limesurvey_sync_api_tablename($this->ls_entity_type, $this->ls_sid, TRUE);
      // Rewrite labels
      $sid_suffix = ' ' . $this->ls_sid;
      $this->label = $this->ls_entity_type . $sid_suffix;
      $this->plural_label = $this->ls_entity_type . 's' . $sid_suffix;
      // Rewrite admin_ui, path.
      $this->crud_path = limesurvey_sync_api_typename($this->ls_entity_type, $this->ls_sid);
      // Use wildcard.
      $survey_type = '%limesurvey_sync_survey'; //str_replace('_', '-', array_search($this->ls_sid, $this->ls_sids));
      $this->root_admin_ui = "admin/structure/limesurvey/surveys/manage/{$survey_type}/" . str_replace('_', '-', $this->crud_path);
      $this->overview_path = $this->root_admin_ui;
      // Do not create the entity if the table does not exists on LS, except
      // the DP table already exists.
      module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
      $this->ls_sids_valid = array();
      foreach ($this->ls_sids as $type => $sid) {
        if (limesurvey_sync_ls_api_ls_table_exists($this->ls_entity_type, $sid) || db_table_exists($this->tablename)) {
          $this->ls_sids_valid[$type] = $sid;
        }
      }
    }
  }

  /** Inform the system about one or more entity types
   *
   * Building entity_info corresponding to settings above.
   *
   *   @see hook_entity_info()
   */
  public function entity_info() {
    $entities_info = array();
    // Create one entity type per survey ID.
    foreach ($this->ls_sids as $type => $sid) {
      $this->ls_sid = $sid;
      $this->set_for_ls_sid();
      if (in_array($sid, $this->ls_sids_valid)) {
        $entities_info = array_merge($entities_info, parent::entity_info());
        // Add custom root ui
        $entities_info[$this->entityname]['crud path'] = limesurvey_sync_api_typename($this->ls_entity_type, $sid);
      }
    }

    return $entities_info;
  }

  /**
   * Define the current version of the database schema (hook_schema()).
   *
   * Building schema corresponding to settings above.
   *
   * @return
   *   An array, the tables schema keyed by the table names.
   *
   *   @see hook_schema()
   */
  public function schema($datas = array()) {
    $schema = array();

    // Create one entity type per survey ID.
    foreach ($this->ls_sids as $type => $sid) {
      $this->ls_sid = $sid;
      $this->set_for_ls_sid();
      if (in_array($sid, $this->ls_sids_valid)) {
        $schema = array_merge($schema, parent::schema(array('ls_sid' => $sid)));
      }
    }

    return $schema;
  }

  protected function schema_dp_table($revision = FALSE, $datas = array()) {
    $schema_table = parent::schema_dp_table($revision);
    if (!empty($data['ls_sid']) && isset($schema_table['description'])) {
      $schema_table['description'] .= " from the LimeSurvey surveys #{$data['ls_sid']}.";
    }
    return $schema_table;
  }

  /**
   * Provide hook_entity_property_info() for the entity.
   *
   * Thoses informations will be used into views, rules, token modules and the
   * entity view page.
   *
   * @return
   *   An array, the entity_property_info() informations.
   *
   *   @see hook_entity_property_info()
   */
  public function entity_property_info($datas = array()) {
    $entity_property_info = array();
    // Create one entity type per survey ID.
    foreach ($this->ls_sids as $type => $sid) {
      $this->ls_sid = $sid;
      $this->set_for_ls_sid();
      if (in_array($sid, $this->ls_sids_valid)) {
        $entity_property_info = array_merge($entity_property_info, parent::entity_property_info(array('ls_sid' => $sid)));
      }
    }

    return $entity_property_info;
  }

  /**
   * Update schema fields regarding changes on the LS database.
   *
   * This method is called on saving a survey entity, compare the current schema
   * stored into a variable and add colomns if needed (as an example : when a
   * question has been added to a survey on the LS admin interface). Set a
   * watchdog entry if add colomns. For the response entity : add corresponding
   * entity reference fields.
   *
   * @return
   *   A boolean, TRUE if the schema has been updated (so the cached schema need
   *   to be rebuild).
   *
   * @see LimesurveySyncInfos::schema()
   * @see LimesurveySyncSurveyController::save()
   */
  public function update_table($datas = array()) {
    // check tables for each survey ID.
    $schema = $this->schema($datas);
    foreach ($this->ls_sids as $type => $sid) {
      $this->ls_sid = $sid;
      $this->set_for_ls_sid();
      if (in_array($sid, $this->ls_sids_valid)) {
        parent::update_table(array('ls_sid' => $sid, 'schema' => $schema));
      }
    }
  }

  /**
   * Provide hook_permission for the entity.
   *
   * @return
   *   An array, the entity permissions.
   *
   * @see hook_permission()
   */
  public function permission($datas = array()) {
    // Set permission for each survey ID.
    $permissions = array();
    foreach ($this->ls_sids as $type => $sid) {
      $this->ls_sid = $sid;
      $this->set_for_ls_sid();
      if (in_array($sid, $this->ls_sids_valid)) {
        // Answers and token contain dynamic author (uid into entity_load()) :
        // act as they own an author colomn for permissions.
        $this->user_col = TRUE;
        $permissions = array_merge($permissions, parent::permission(array('ls_sid' => $sid)));
      }
    }
    return $permissions;
  }
}

/**
 * Class managing entities.
 *
 * Extend the common class 'Entity' for entities.
 */
class LimesurveySyncEntityLs extends LimesurveySyncEntity {
  public $ls_entity_type = NULL;

  /**
   * Creates a new entity.
   *
   * Retreive the LS survey ID from the entity type name, making sure to find it
   * for answers and token entities.
   *
   * @see entity_create()
   */
  public function __construct(array $values = array(), $entityType = NULL) {

    $with_suffix = array('token', 'answer');
    if (in_array($this->ls_entity_type, $with_suffix)) {
      $ls_sid = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entityType, $this->ls_entity_type);
      if (empty($ls_sid)) {
        throw new Exception('Cannot create an instance of ' . $this->ls_entity_type . ' entity without a specified survey id.');
      }
      else {
        $this->ls_sid = $ls_sid;
      }
    }
    parent::__construct($values, $entityType);
  }

  /**
   * Synchronize the entity from LS to DP .
   *
   * @param $all
   *   A string, 'all' or 'all_skip_import' for synchronizing all correspopnding
   *   entities, or FALSE for synchronizing the provided entity. Default to
   *   FALSE.
   */
  public function ls_synchronize($all = FALSE, $way = 'ls2dp') {
    $arg = ($all != 'all') ? $this : 'all';
    return entity_get_controller($this->entityType)->ls_synchronize($arg, $way);
  }

  /**
   * Returns whether the survey type is locked, thus may not be deleted or renamed.
   *
   */
  public function isLocked() {

    return isset($this->status) && empty($this->is_new) && (($this->status & ENTITY_IN_CODE) || ($this->status & ENTITY_FIXED));
  }
}

/**
 *  A controller implementing exportables stored in the database.
 *
 * Extend the common class 'EntityAPIControllerExportable' for entities.
 */
class LimesurveySyncController extends LimesurveySyncEntityController {
  public $ls_entity_type = '';
  public $ls_sid = '';
  public $ls_key = '';

  /**
   * Overridden.
   *
   * Retreive the LS survey ID from the entity type name, making sure to find it
   * for answers and token entities.
   */
  public function __construct($entityType) {

    $with_suffix = array('token', 'answer');
    $ls_sid = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entityType, $this->ls_entity_type);
    if (empty($ls_sid) && in_array($this->ls_entity_type, $with_suffix)) {
      throw new Exception('Cannot create an instance of ' . $this->ls_entity_type . ' entity without a specified survey id.');
    }
    else {
      $this->ls_sid = $ls_sid;
    }
    parent::__construct($entityType);
  }

  /**
   * Overridden.
   *
   * Set default values on creating entity.
   */
  public function create(array $values = array()) {
    // Add values that are specific to our LimesurveySyncSurvey
    $entity_infos = limesurvey_sync_api_info_class_object($this->ls_entity_type);
    $ls_sid = (!empty($values['ls_sid'])) ? $values['ls_sid'] : $this->ls_sid;
    $fields = $entity_infos->entity_default_values($ls_sid);
    $values += $fields;
    $entity = parent::create($values);
    return $entity;
  }

  /**
   * Overridden.
   *
   * Add dynamic title for answers and token entities.
   */
  public function load($ids = array(), $conditions = array()) {

    $entities = parent::load($ids, $conditions);
    if (empty($this->entityInfo['entity keys']['label']) && !empty($this->entityInfo['entity keys']['ls key'])) {
      // For answers and token entities.
      $ls_key = $this->entityInfo['entity keys']['ls key'];
      foreach ($entities as $entity) {
        // Add dynamic title.
        if (!isset($entity->title)) {
          $entity->title = limesurvey_sync_api_default_label($entity, $ls_key);
        }

        // Provide uid from the response table.
        // Note : uid may be an array if various values exists. Pick the first one corresponding to the most recent
        if (!isset($entity->uid)) {
          $uid = limesurvey_sync_api_retreive_uid_from_ref_entity($entity);
          $entity->uid = (!is_array($uid)) ? $uid : current($uid);
        }
      }
    }
    return $entities;
  }

  /**
   * Provide the LimeSurvey Sync module entity status.
   *
   * Corresponding to LS states (editing, completed, etc...). each subclass will
   * implement it.
   *
   * @param $entity
   *   The entity object.
   *
   * @return
   *   A numeric defined as a constant corresponding to a defined LS stage.
   */
  protected function lss_status($entity) {
    // Each class will implement it.
    return FALSE;
  }

  /**
   * Provide the LimeSurvey Sync module entity synchronzation status.
   *
   * Check if the Drupal entity is binded to a valid LimeSurvey entry.
   *
   * @param $entity
   *   The entity object.
   * @param $existing_ls_enties_keyed_by_ls_key
   *   An array : the LimeSurvey entries keyed by the LS primary key. For
   *   performance raisons, this parameter will usually contain only one key
   *   corresponding of the one from the DP entry. Default is empty.
   *
   * @return
   *   A numeric defined as a constant : LIMESURVEY_SYNC_SYNCHRONIZED (1) for
   *   synchronized, LIMESURVEY_SYNC_DESYNCHRONIZED (2) for desynchronized (this
   *   entry had exist into the ls table but has been deleted into the ls table)
   *   or LIMESURVEY_SYNC_UNSYNCHRONIZED (3) for unsynchronized (this dp entry
   *   has never had a reference into the ls table).
   */
  protected function lss_sync($entity, $existing_ls_enties_keyed_by_ls_key = array()) {

    $ls_key = isset($this->entityInfo['entity keys']['ls key']) ? $this->entityInfo['entity keys']['ls key'] : '';
    if (!$ls_key || empty($entity->$ls_key)) {
      // This dp entry has never had a reference into the ls table. This
      // situation should not exists.
      return LIMESURVEY_SYNC_UNSYNCHRONIZED;
    }
    elseif (array_key_exists($entity->$ls_key, $existing_ls_enties_keyed_by_ls_key)) {
      return LIMESURVEY_SYNC_SYNCHRONIZED;
    }
    else {
      // This entry had exist into the ls table but has been deleted into
      // the ls table.
      return LIMESURVEY_SYNC_DESYNCHRONIZED;
    }
  }

  /**
   * Permanently saves the given entity, synchronizing it.
   *
   * On saving entity, update/add informations from LimeSurvey to Drupal. Reset
   * the flag 'need_save_ls2dp' to FALSE. Manage date user, log fields. In case
   * of failures, an exception is thrown.
   *
   * @param $entity
   *   The entity to save.
   *
   * @return
   *   SAVED_NEW or SAVED_UPDATED is returned depending on the operation
   *   performed.
   */
  public function save($entity, DatabaseTransaction $transaction = NULL) {

    // Entity can be saved using entity_save(). In order to avoid calling it
    // twice and to synchronize datas on saving, add a flag.
    // Revisions should not be synchronized.
    $is_default_revision = entity_revision_is_default($this->entityType, $entity);
    $need_save_ls2dp = (!isset($entity->need_save_ls2dp) || $entity->need_save_ls2dp);
    $not_activate_token = (!isset($entity->activate_token) || !$entity->activate_token);
    $need_save_dp2ls = (isset($entity->need_save_dp2ls) && $entity->need_save_dp2ls);
    if ($is_default_revision) {
      $dp_key = $this->idKey;
      $dp_key_value = (!empty($entity->$dp_key)) ? $entity->$dp_key : 'this';
      if ($need_save_dp2ls) {
        $synchronize_set = $this->ls_synchronize($entity, 'dp2ls');
        // If synchronization needed : apply the changes.
        if (!empty($synchronize_set['need_sync'])) {
          $diff = $synchronize_set['need_sync'][$dp_key_value]['diff'];
          // Update the LS entry.
          module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
          $succeed = limesurvey_sync_ls_api_edit_row($entity, $diff);
        }
        if ($succeed) {
          // Reset the flag.
          $entity->need_save_dp2ls = FALSE;
          // Save revision on each changes on LS.
          $entity->revision = TRUE;
          // Set a log message expliciting diff fields.
          $entity->log = t('Sync DP -> LS : changes on @fields.', array('@fields' => implode(', ', array_keys($diff))));
        }

        // Sync ls2dp now for lss_status and lss_sync.
        $need_save_ls2dp = TRUE;
      }

      if ($need_save_ls2dp && $not_activate_token) {

        $synchronize_set = $this->ls_synchronize($entity);

        // If synchronization needed : apply the changes.
        if (!empty($synchronize_set['need_sync'])) {
          $diff = $synchronize_set['need_sync'][$dp_key_value]['diff'];
          foreach ($diff as $key => $value) {
            $entity->$key = $value;
          }
          // Reset the flag.
          $entity->need_save_ls2dp = FALSE;
          // Save revision on each changes on LS.
          $entity->revision = TRUE;
          // Set a log message expliciting diff fields.
          $entity->log = t('Sync LS -> DP : changes on @fields.', array('@fields' => implode(', ', array_keys($diff))));
        }
      }
    }

    // Use entity_revision_set_default($entity_type, $entity) for make-current.

    $return = parent::save($entity, $transaction);

  }

  /**
   * Test the synchronization from LimeSurvey to Drupal or Dp2LS..
   *
   * Compare the DP entry to the current LS one. Update the entity adding a flag
   * 'need_save_ls2dp' to TRUE or FALSE. This function can be called for one
   * entity or for all entities from this entity type.
   *
   * @param $entity
   *   The entity object to save or 'all' for all entities from this entity
   *   type, or 'all_skip_import' for all entities from this entity
   *   type but without returning (for permformance raisons) entities that
   *   have never been saved on DP.
   * @param $way
   *   A string, the synchronization way. 'ls2dp' or 'dp2ls'. Default is 'ls2dp'.
   *
   * @return
   *   An array. Keys are ['ok'][dp_primary_key_value] if synchronization is
   *   not needed; ['need_sync'] if synchronization is needed, with subkeys :
   *   ['need_sync'][dp_primary_key_value]['dp'] the DP current entity object,
   *   ['need_sync'][dp_primary_key_value]['diff'] an array of the changes keyed
   *   by the properties; ['need_sync'][dp_primary_key_value]['dp_date'] the
   *   last changed date on DP and ['need_sync'][dp_primary_key_value]['ls_date']
   *   the last changed date on LS (used to compare witch one is the most
   *   recent, for selecting ls2dp or dp2ls synchronization in future
   *   developpements). At last ['need_import'][ls_primary_key_value] as the
   *   converted to DP entity object to save for LS entries that have no entry
   *   in DP. ['need_import'] only exists with the 'all' parameter. Note : for
   *   entity to create, the dp_primary_key_value from
   *   ['need_sync'][dp_primary_key_value] is 'this'. It should be taken in
   *   consideration when browsing the result of the ls_synchronize()
   *   method.
   */
  public function ls_synchronize(&$entity, $way = 'ls2dp') {

    $dp_key = $this->idKey; // entityInfo['entity keys']['id'];
    $ls_key = isset($this->entityInfo['entity keys']['ls key']) ? $this->entityInfo['entity keys']['ls key'] : '';
    $this_ls_id = FALSE;
    $this_dp_id = FALSE;

    // Retrieve from Drupal.
    if ($entity != 'all' && $entity != 'all_skip_import') {
      // If an antity is provided sync only this one.
      // The entity may have been altered : use this one.
      $entities = array($entity);
      if (!empty($entity->$dp_key)) {
        $this_dp_id = $entity->$dp_key;
      }
      else {
        // Only one entity : Use 'this' index.
        $this_dp_id = 'this';
      }

      // For retreiving from LS.
      if ($ls_key && !empty($entity->$ls_key)) {
        $this_ls_id = $entity->$ls_key;
      }
      else { //if (!$ls_key) {
        // The response entity arrive here.
        // Try to create/update an entity with no ID references to LS.
        // Note : To add a LS token, save it first on LS, then synchronize it
        // using its new ID.
      }
    }
    else {
      // Sync all entities.
      // Retrieve from Drupal.
      $entities = entity_load($this->entityType);
      ksort($entities);
    }

    $synchronize_set = array();
    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_fields');
    // Retrieve form LS.
    if ($this->ls_entity_type == 'response') {
      $ls_ids = array();
    }
    elseif ($this->ls_entity_type == 'survey') {
      module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_survey');
      $all_sids = limesurvey_sync_survey_properties('all');
      $ls_ids = ($all_sids) ? $all_sids : array();
    }
    else {
      if ($entity == 'all_skip_import' && $ls_key) {
        $selsected_ls_ids = array();
        foreach ($entities as $record) {
          $selsected_ls_ids[] = $record->$ls_key;
        }
        if (!empty($selsected_ls_ids)) {
          $this_ls_id = $selsected_ls_ids;
        }
      }
      $ls_rows = limesurvey_sync_ls_api_retrieve_ls_row($this->ls_entity_type, $this->ls_sid, $this_ls_id);
      $ls_ids = ($ls_rows) ? $ls_rows : array();
    }
    $synchronize_set['need_import'] = $ls_ids;

    foreach ($entities as $record) {
      $dp_key_value = ($this_dp_id != 'this') ? $record->$dp_key : 'this';
      // Break the association object. Work on array.
      $dp_row = (array) $record;
      $ls_row_converted = array(
        'lss_status' => $this->lss_status($record),
        'lss_sync' => $this->lss_sync($record, $ls_ids),
      );
      // Don't change the values for LS fields if not synchronized.
      if ($ls_row_converted['lss_sync'] == LIMESURVEY_SYNC_SYNCHRONIZED) {
        // This dp entry does exists in both dp and ls : an update may be needed.
        // On the survey and response entity, only sync and status are concerned.
        if (!in_array($this->ls_entity_type, array('survey', 'response'))) {
          $ls_row = $ls_ids[$record->$ls_key];
          // Convert LS fields to drupal.
          $ls_row_converted = array_merge($ls_row_converted, limesurvey_sync_ls_api_fields_convert_values($ls_row, $this->ls_entity_type, $this->ls_sid, 'ls2dp'));
          // And update status values.
          $ls_row_converted['lss_status'] = $this->lss_status($ls_row_converted);
        }
        // Delete entries that have been handled.
        if ($ls_key) {
          unset($synchronize_set['need_import'][$record->$ls_key]);
        }
      }
      // @TODO : for dp2ls on all entities, look for desynchronized entities,
      // proposing to recreate them on LS.
      // Compare both.
      if ($way == 'ls2dp') {
        $diff = array_diff_assoc($ls_row_converted, $dp_row);
      }
      else {
        $ls_fields = array_flip(limesurvey_sync_ls_api_retrieve_ls_fields($entity->ls_entity_type, $entity->ls_sid));
        // Select into $tochange the LS fields.
        $dp_row_ls_fields = array_intersect_key($dp_row, $ls_fields);
        $diff = array_diff_assoc($dp_row_ls_fields, $ls_row_converted);
      }
      if (!empty($diff)) {
        // The entity need to be resync and save !
        $dp_row_sync = $dp_row;
        foreach ($diff as $field => $value) {
          $dp_row_sync[$field] = $value;
        }
        if (isset($this->entityInfo['entity keys']['revision'])) {
          $dp_row_sync['log'] = format_plural(count($diff), 'Change on the field %fields', 'Change on the fields : %fields', array(
            '%fields' => implode(', ', array_keys($diff)),
          ));
          $dp_row_sync['revision'] = TRUE;
        }

        // Save the result into an object. Use entity_save() on it to save it.
        $synchronize_set['need_sync'][$dp_key_value]['dp'] = $record;
        $synchronize_set['need_sync'][$dp_key_value]['diff'] = $diff;
        // Save Dates, it may have an interest for saving the most recent one.
        if (!empty($this->entityInfo['entity keys']['ls date key']) && !empty($dp_row_sync['changed'])) {
          // No date for survey.
          $synchronize_set['need_sync'][$dp_key_value]['dp_date'] = $dp_row_sync['changed'];
          $synchronize_set['need_sync'][$dp_key_value]['ls_date'] = $ls_row_converted[$this->entityInfo['entity keys']['ls date key']];
        }
        // if it match the entity requested : update it with the new values.
        if ($entity != 'all' && $this_dp_id && $this_dp_id == $dp_key_value) {
          // $entity is provided by reference. update it with a flag.
          $entity->{"need_save_$way"} = TRUE;
        }
      }
      else {
        // If no change, do not store them for performance raisons.
        $synchronize_set['ok'][$dp_key_value] = TRUE;
        // if it match the entity requested : set a no sync flag..
        if ($entity != 'all' && $this_dp_id && $this_dp_id == $dp_key_value) {
          // $entity is provided by reference. Update it.
          $entity->{"need_save_$way"} = FALSE;
        }
      }
    }

    // All entities from Drupal have been explored. Look for remaining entities
    // to import from LS.
    // Don't try to import survey : it only manually.
    if ($way == 'ls2dp' && !empty($synchronize_set['need_import']) && $this->ls_entity_type != 'survey') {
      foreach ($synchronize_set['need_import'] as $ls_id => $ls_row) {
        $ls_row_converted = limesurvey_sync_ls_api_fields_convert_values($ls_row, $this->ls_entity_type, $this->ls_sid, 'ls2dp');
        $ls_row_converted['lss_sync'] = LIMESURVEY_SYNC_SYNCHRONIZED;
        $ls_row_converted['lss_status'] = $this->lss_status($ls_row_converted);
        $entity_to_import = entity_create($this->entityType, $ls_row_converted);
        // Add a flag to not synchronize twice from the save() method.
        $entity_to_import->need_save_ls2dp = TRUE;
        // Remplace into the synchronize_set et LS array by the new entity
        // object. It can be created using entity_save().
        $synchronize_set['need_import'][$ls_id] = $entity_to_import;
        if ($entity != 'all' && $ls_id == $this_ls_id) {
          // $entity is provided by reference. update it with a flag.
          //$entity = $synchronize_set['need_import'][$ls_id];
          $entity->need_save_ls2dp = TRUE;
        }
      }
    }

    if (empty($synchronize_set['need_import'])) {
      unset($synchronize_set['need_import']);
    }

    return $synchronize_set;
  }

  /**
   * Builds a structured array representing the entity's content.
   *
   * The content built for the entity will vary depending on the $view_mode
   * parameter.
   *
   * @param $entity
   *   An entity object.
   * @param $view_mode
   *   View mode, e.g. 'full', 'teaser'...
   * @param $langcode
   *   (optional) A language code to use for rendering. Defaults to the global
   *   content language of the current request.
   *
   * @return
   *   The renderable array.
   */
  public function buildContent($entity, $view_mode = 'full', $langcode = NULL, $content = array()) {

    // Our additions to the $build render array
    // $wrapper = entity_metadata_wrapper($this->entityType, $entity);
    $build = parent::buildContent($entity, $view_mode, $langcode, $content);

    return $build;
  }
}

/**
 * Default controller for providing UI.
 *
 * Extend the common class 'EntityDefaultUIController' for entities.
 */
class LimesurveySyncUIController extends LimesurveySyncEntityUIController {
  // Edit the menu from EntityDefaultUIController::hook_menu().
  public $ls_entity_type = '';
  public $ls_sid = '';
  protected $menu_weight = 0;
  protected $ls_sub_entity = FALSE;
  protected $hide_export_status = TRUE;

  // Config : For now I disallow to edit and create answer and token.
  // I may change my mind on this later.
  protected $disallow_add = FALSE;
  protected $disallow_edit = FALSE;
  // response overview and add link are into the 'content' root, other are into
  // 'admin' root.

  /**
   * Overridden.
   *
   * Retreive the LS survey ID from the entity type name, making sure to find it
   * for answers and token entities.
   */
  public function __construct($entity_type, $entity_info) {

    $with_suffix = array('token', 'answer');
    $ls_sid = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entity_type, $this->ls_entity_type);
    if (empty($ls_sid) && in_array($this->ls_entity_type, $with_suffix)) {
      throw new Exception('Cannot create an instance of ' . $this->ls_entity_type . ' entity without a specified survey id.');
    }
    else {
      $this->ls_sid = $ls_sid;
    }
    parent::__construct($entity_type, $entity_info);
  }

  /**
   * Provides definitions for implementing hook_menu().
   *
   * Manage entity view page, adding, synchronize (one and all) pages, revisions,
   * import, clone, devel pages.
   */

  public function hook_menu() {

    // In order to works with the menu router, the modules must be called in
    // this order : entity, limesurvey_sync, limesurvey_sync_survey, the answer
    // and token, depending on how the hook_entity_info() functions are called.
    $items = parent::hook_menu();
    // Ad specific stuf for sub entities.
    if ($this->ls_sub_entity) {
      // Set token and answer list as tab instead of callback.
      $overview_path = $this->entityInfo['overview path'];
      if (isset($items[$overview_path])) {
        $items[$overview_path]['type'] = MENU_LOCAL_TASK;
        $items[$overview_path]['weight'] = $this->menu_weight;
        $items[$overview_path]['access callback'] = 'limesurvey_sync_api_subentity_tab_access';
        $items[$overview_path]['access arguments'] = array('update', $this->entityType, (count(explode('/', $this->path)) -2));
      }
    }
    // Cloning a entry may be a problem for linking drupal and limesurvey.
    // Disalow it !
    $wildcard = isset($this->entityInfo['admin ui']['menu wildcard']) ? $this->entityInfo['admin ui']['menu wildcard'] : '%entity_object';
    $path_view = "{$this->entityInfo['crud path']}/{$wildcard}";
    $path_clone = $path_view . '/clone';
    if (isset($items[$path_clone])) {
      // I keep it untill I 'm sure it's a trouble.
      //unset($items[$path_clone]);
    }
    if ($this->ls_entity_type ==  'survey') {
      // Adapt the survey view page for the menu : not menu callback.
      $items[$path_view]['type'] = MENU_NORMAL_ITEM;
    }

    // Add synchronize link.
    $id_count = count(explode('/', $path_view));
    $placeholder_view = $id_count - 1;

    // Set a synchronization page for all entities.
    $items[$this->entityInfo['overview path'] . '/synchronize-all'] = array(
      'title' => 'Synchronize all ' . $this->ls_entity_type,
      'page callback' => 'drupal_get_form',
      'page arguments' => array(
        'limesurvey_sync_api_synchronize_page',
        $this->entityType,
        FALSE,
      ),
      'access arguments' => array("administer limesurvey sync {$this->ls_entity_type}"),
      'weight' => -5,
      'type' => MENU_LOCAL_ACTION,
      'file' => 'limesurvey_sync_api.pages.inc',
      'file path' => drupal_get_path('module', 'limesurvey_sync') . '/api',
    );

    // Set a synchronization page for 1 entity.
    $items[$path_view . '/synchronize'] = array(
      'title' => 'Synchronize',
      // 'title callback' => 'entity_label',
      // 'title arguments' => array($this->entityType, $id_count + 1),
      'page callback' => 'drupal_get_form',
      'page arguments' => array(
        'limesurvey_sync_api_synchronize_page',
        $this->entityType,
        $placeholder_view,
      ),
      'load arguments' => array($this->entityType),
      'access callback' => 'entity_access',
      'access arguments' => array(
        'update',
        $this->entityType,
        $placeholder_view,
      ),
      'type' => MENU_LOCAL_TASK,
      'file' => 'limesurvey_sync_api.pages.inc',
      'file path' => drupal_get_path('module', 'limesurvey_sync') . '/api',
    );

    // Mofifie revert revision callback
    if (isset($items[$path_view . '/revisions/%/revert'])) {
      $items[$path_view . '/revisions/%/revert']['page arguments'][0] = 'limesurvey_sync_api_revision_revert_confirm';
      $items[$path_view . '/revisions/%/revert']['file'] = 'limesurvey_sync_api.pages.inc';
    }

    return $items;
  }

  /**
   * Provides definitions for implementing hook_forms().
   *
   * Use per bundle form ids if possible, such that easy per bundle alterations
   * are supported too.
   *
   * Manage editing form for answers and tokens entities. Note that for
   * performance reasons, this method is only invoked for forms, which receive
   * the entity_type as first argument. Thus any forms added, must follow that
   * pattern.
   *
   * @see entity_forms()
   */
  public function hook_forms() {

    $forms = parent::hook_forms();
    if ($this->ls_sub_entity) {
      foreach (limesurvey_sync_survey_get_surveys() as $type) {
        $entity_type = limesurvey_sync_api_entityname($this->ls_entity_type, $type->ls_sid);
        $form_id_value = $entity_type . '_form';
        $forms[$form_id_value] = array(
          'callback' => "limesurvey_sync_{$this->ls_entity_type}_form",
          //'wrapper_callback' => 'entity_ui_form_defaults',
        );
      }
    }

    return $forms;
  }

  /**
   * Generates the table headers for the overview table.
   */
  protected function overviewTableHeaders($conditions, $rows, $additional_header = array()) {

    $additional_header[] = t('synchronization status');
    $additional_header[] = t('LS status');

    return parent::overviewTableHeaders($conditions, $rows, $additional_header);
  }

  /**
   * Generates the row for the passed entity and may be overridden in order to
   * customize the rows.
   *
   * Add the synchronization status and LimeSurvey status colomns and the
   * synchronize operation. Delete export status colomn and edit operation
   * depending on settings ($hide_export_status and $disallow_edit).
   *
   * @param $additional_cols
   *   Additional columns to be added after the entity label column.
   */
  protected function overviewTableRow($conditions, $id, $entity, $additional_cols = array(), $additional_ops = array()) {

    // Disable stuff.
    //return parent::overviewTableRow($conditions, $id, $entity, $additional_cols);
    // Rewrite te whole method, since all links changes.
    if (!is_numeric($id)) {
      $id = str_replace('_', '-', $id);
    }
    static $synchronize_set = array();
    if (empty($synchronize_set)) {
      // HERE : retrieve the synchronization set. Call it only one time.
      $entity_type = $this->entityType;
      $all = 'all_skip_import'; //'all'
      $synchronize_set = entity_get_controller($entity_type)->ls_synchronize($all);
    }
    $dp_key = $this->entityInfo['entity keys']['id'];
    $css = array(
      'css' => array(
        drupal_get_path('module', 'limesurvey_sync') . '/misc/limesurvey_sync.css',
      ),
    );

    $additional_cols = array();

    //$sync_entity = $entity;
    $need_sync = FALSE;
    $need_sync_entity = (!empty($synchronize_set['need_sync']) && array_key_exists($entity->$dp_key, $synchronize_set['need_sync']));
    $need_sync_ref = (!empty($synchronize_set['need_sync_ref']) && array_key_exists($entity->$dp_key, $synchronize_set['need_sync_ref']));
    $lss_sync = $entity->lss_sync;
    $lss_status = $entity->lss_status;
    if ($need_sync_entity || $need_sync_ref) {
      $need_sync = TRUE;
      // Build the synchronized entity.
      if ($need_sync_entity) {
        //$sync_entity = $synchronize_set['need_sync'][$entity->$dp_key]['dp'];
        if (isset($synchronize_set['need_sync'][$entity->$dp_key]['diff']['lss_sync'])) {
          $lss_sync = $synchronize_set['need_sync'][$entity->$dp_key]['diff']['lss_sync'];
        }
        if (isset($synchronize_set['need_sync'][$entity->$dp_key]['diff']['lss_status'])) {
          $lss_status = $synchronize_set['need_sync'][$entity->$dp_key]['diff']['lss_status'];
        }
        //foreach ($synchronize_set['need_sync'][$entity->$dp_key]['diff'] as $key => $value) {
//  $sync_entity->$key = $value;
        //}
      }
      else {
        //$sync_entity = $synchronize_set['need_sync_ref'][$entity->$dp_key]['dp'];
      }
    }
    // lss_sync col. Display the sync status or 'To synchronize' if sync needed.
    $additional_cols[] = array(
      'data' => array(
        '#theme' => 'limesurvey_sync_field_label',
        '#field' => 'lss_sync',
        '#to_display' => $lss_sync,
        '#ls_entity_type' => $entity->ls_entity_type,
        '#ls_sid' => $entity->ls_sid,
        '#need_sync' => $need_sync,
        '#attached' => $css,
      ),
    );
    // lss_status_col.
    $additional_cols[] = array(
      'data' => array(
        '#theme' => 'limesurvey_sync_field_label',
        '#field' => 'lss_status',
        '#to_display' => $lss_status,
        '#ls_entity_type' => $entity->ls_entity_type,
        '#ls_sid' => $entity->ls_sid,
        '#attached' => $css,
      ),
    );

    // Synchronize link col : it will be included into the operation header.
    $path_view = $this->entityInfo['crud path'];
    $additional_ops = array();
    $additional_ops[] = l(t('synchronize'), $path_view . '/' . $id . '/synchronize');
    return parent::overviewTableRow($conditions, $id, $entity, $additional_cols, $additional_ops);
  }

  /**
   * Returns the operation count for calculating colspans.
   */
  protected function operationCount() {

    $count = parent::operationCount();
    // The synchronize operation.
    $count += 1;
    return $count;
  }

}

/**
 * Sets the property to the given value. May be used as 'setter callback'.
 */
function limesurvey_sync_api_setter_callback($data, array $options, $name, $type, $info) {
  return entity_property_verbatim_set($data, $options, $name, $type, $info);
}


/**
 * Determines whether the given user has access to an entity.
 *
 * @param $op
 *   The operation being performed. One of 'view', 'update', 'create' or
 *   'delete'.
 * @param $entity_type
 *   The entity type of the entity to check for.
 * @param $entity
 *   Optionally an entity to check access for. If no entity is given, it will be
 *   determined whether access is allowed for all entities of the given type.
 * @param $account
 *   The user to check for. Leave it to NULL to check for the global user.
 *
 * @return boolean
 *   Whether access is allowed or not. If the entity type does not specify any
 *   access information, NULL is returned.
 *
 * @see entity_type_supports()
 */
function limesurvey_sync_api_entity_access($op, $entity = NULL, $account = NULL, $entity_type = NULL) {
  // Map the op value. EntityDefaultUIController:hook_menu() use 'add' and
  // 'update'. entity_access() will work even if 'create' or 'edit' are provided.
  switch ($op) {
    case 'edit':
      $op = 'update';
      break;
    case 'create':
      $op = 'add';
      break;
  }

  // Now op and entity_type are set : check them.
  // First retreive the ls_entity_type from the entity type name, as $entity may be null.
  if ($entity_type && !($ls_entity_type = limesurvey_sync_api_retrieve_ls_entity_type_from_entityname($entity_type))) {
    // In the menu tree, make sure the answers and token entities are displayed
    // into the same menu tree that the corresponding survey.
    return FALSE;
  }

  // @TODO: add 'synchronize' op.
  if ((empty($entity) && empty($entity_type)) || !in_array($op, array('view', 'update', 'delete', 'add'), TRUE)) {
    // If there was no entity to check against, or the $op was not one of the
    // supported ones, we return access denied.
    return FALSE;
  }

  // If no user object is supplied, the access check is for the current user.
  if (empty($account)) {
    $account = $GLOBALS['user'];
  }

  // Check for high level permission.
  if (user_access("administer limesurvey sync {$ls_entity_type}", $account)) {
    return TRUE;
  }

  if ($ls_entity_type == 'response' && $op == 'add' && (empty($entity) || empty($entity->ls_sid))) {
    // This situation should not exist : for checking response creation
    // permission, the $entity must be provided as an object. That why I create
    // the _limesurvey_sync_api_entity_access() function.
    return FALSE;
  }

  // Prepare the user_access() string.
  $no_bundles = (in_array($ls_entity_type, array('survey', 'participant')));
  if (!$no_bundles) {
    // ls_sid needed.
    $ls_sid = NULL;
    $ls_sid_from_entitytype = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entity_type, $ls_entity_type);
    if (!empty($entity->ls_sid)) {
      $ls_sid = $entity->ls_sid;
    }
    elseif ($ls_sid_from_entitytype) {
      $ls_sid = $ls_sid_from_entitytype;
    }
    else {
      $entity = ($ls_entity_type != 'response') ? entity_create($entity_type, array()) : NULL;
      if (!empty($entity->ls_sid)) {
        $ls_sid = $entity->ls_sid;
      }
    }

    if ($ls_sid) {
      $survey_type = check_plain(limesurvey_sync_api_typename('survey', $ls_sid));
      if ($survey_type) {
        $perm_machine_suffix = "{$survey_type} {$ls_entity_type}";
      }
      else {
        return FALSE;
      }
    }
    else {
      return FALSE;
    }
  }
  else {
    $perm_machine_suffix = $ls_entity_type;
  }

  // Handle operations.
  // Create entity.
  if ($op == 'add') {
    if (user_access("$op {$perm_machine_suffix}", $account)) {
      return TRUE;
    }
    return FALSE;
  }
  // View, delete, update entity.
  if (user_access("$op any {$perm_machine_suffix}", $account)) {
    return TRUE;
  }
  if (user_access("$op own {$perm_machine_suffix}", $account) && isset($entity->uid) && $entity->uid == $account->uid) {
    return TRUE;
  }

  // At last.
  return FALSE;
}

/**
 * Determines whether the given user has access to an entity.
 *
 * @param $op
 *   The operation being performed. One of 'view', 'update', 'create' or
 *   'delete'.
 * @param $entity_type
 *   The entity type of the entity to check for.
 * @param $entity
 *   Optionally an entity to check access for. If no entity is given, it will be
 *   determined whether access is allowed for all entities of the given type.
 * @param $account
 *   The user to check for. Leave it to NULL to check for the global user.
 *
 * @return boolean
 *   Whether access is allowed or not. If the entity type does not specify any
 *   access information, NULL is returned.
 *
 * @see entity_type_supports()
 */
function limesurvey_sync_api_subentity_tab_access($op, $entity_type, $survey, $entity = NULL, $account = NULL ) {
  if (!is_object($survey)) {
    return FALSE;
  }
  // Check if the survey and the entity type match.
  $sub_entity_ls_entity_type = limesurvey_sync_api_retrieve_ls_entity_type_from_entityname($entity_type);
  $sub_entity_ls_sid = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entity_type, $sub_entity_ls_entity_type);
  if ($sub_entity_ls_sid != $survey->ls_sid) {
    // In the menu tree, make sure the answers and token entities are displayed
    // into the same menu tree that the corresponding survey.
    return FALSE;
  }
  else {
    return entity_access($op, $entity_type, $entity, $account);
  }
}

/**
 * Title for token and answser is not stored int Dabase but dynamically loaded.
 *
 * @param $entity_type
 *   The entity type of the entity to check for.
 * @param $entity
 *   Optionally an entity to check access for. If no entity is given, it will be
 *   determined whether access is allowed for all entities of the given type.
 *
 * @return string
 *   a sanitized title.
 *
 * @see entity_label()
 */
function limesurvey_sync_api_sub_label_callback($entity, $entity_type) {
  if (!empty($entity->title)) {
    return $entity->title;
  }
  return '';
}

/**
 * Access callback: Checks entity revision access.
 *
 * Inspired from _node_revision_access().
 *
 * @param $entity_type
 *   The entity type of the entity to check for.
 * @param $entity
 *   Optionally an entity to check access for. If no entity is given, it will be
 *   determined whether access is allowed for all entities of the given type.
 * @param $op
 *   (optional) The specific operation being checked. Defaults to 'view.'
 * @param object $account
 *   (optional) A user object representing the user for whom the operation is
 *   to be performed. Determines access for a user other than the current user.
 *
 * @return
 *   TRUE if the operation may be performed, FALSE otherwise.
 */
function _limesurvey_sync_api_revision_access($entity_type, $entity = NULL, $op = 'view', $account = NULL) {

  if (is_numeric($entity)) {
    $entity = entity_revision_load($entity_type, $entity);
  }

  if (!($ls_entity_type = limesurvey_sync_api_retrieve_ls_entity_type_from_entityname($entity_type))) {
    return FALSE;
  }
  if (!($ls_sid = limesurvey_sync_api_retrieve_ls_sid_from_entityname($entity_type, $ls_entity_type))) {
    return FALSE;
  }

  // Find survey.
  $survey_type = limesurvey_sync_api_typename('survey', $ls_sid);
  if (!($survey = limesurvey_sync_survey_get_surveys($survey_type))) {
    return FALSE;
  }

  $access = &drupal_static(__FUNCTION__, array());

  $map = array(
    'view' => "view {$survey_type} {$ls_entity_type} revisions",
    'update' => "revert {$survey_type} {$ls_entity_type} revisions",
    'delete' => "delete {$survey_type} {$ls_entity_type} revisions",
  );

  if (!$entity || !isset($map[$op])) {
    // If there was no entity to check against, or the $op was not one of the
    // supported ones, we return access denied.
    return FALSE;
  }

  if (!isset($account)) {
    $account = $GLOBALS['user'];
  }

  // Statically cache access by revision ID, user account ID, and operation.
  $cid = $entity->vid . ':' . $account->uid . ':' . $op;

  if (!isset($access[$cid])) {
    // Perform basic permission checks first.
    if (!user_access($map[$op], $account) && !user_access("administer limesurvey sync {$ls_entity_type}", $account)) {
      return $access[$cid] = FALSE;
    }

    $entity_info = entity_get_info($entity_type);
    $id_field = $entity_info['entity keys']['id'];
    $entity_current_revisions = entity_load($entity_type, array($entity->$id_field));
    $entity_current_revision = $entity_current_revisions[$entity->$id_field];
    $is_current_revision = entity_revision_is_default($entity_type, $entity);

    // There should be at least two revisions. If the vid of the given entity and
    // the vid of the current revision differ, then we already have two
    // different revisions so there is no need for a separate database check.
    // Also, if you try to revert to or delete the current revision, that's not
    // good.

    if ($is_current_revision && (db_query('SELECT COUNT(' . $entity_info['entity keys']['revision'] . ') FROM {' . $entity_info['revision table'] . '} WHERE ' . $id_field . ' = :' . $id_field, array(
      ":{$id_field}" => $entity->$id_field,
    ))->fetchField() == 1 || $op == 'update' || $op == 'delete')) {
      $access[$cid] = FALSE;
    }
    elseif (user_access("administer limesurvey sync {$ls_entity_type}", $account)) {
      $access[$cid] = TRUE;
    }
    else {
      // First check the access to the current revision and finally, if the entity
      // passed in is not the current revision then access to that, too.
      $access[$cid] = entity_access($op, $entity_type, $entity_current_revision, $account) && ($is_current_revision || entity_access($op, $entity_type, $entity, $account));
    }

    return $access[$cid];
  }
}

/**
 * Generates an array which displays a node detail page.
 *
 * Inspired from node_show().
 *
 * @param $entity_type
 *   The entity type of the entity to check for.
 * @param $entity
 *   Optionally an entity to check access for. If no entity is given, it will be
 *   determined whether access is allowed for all entities of the given type.
 * @param $message
 *   A flag which sets a page title relevant to the revision being viewed.
 *
 * @return
 *   A $page element suitable for use by drupal_render().
 */
function limesurvey_sync_api_show($entity_type, $entity, $message = FALSE) {

  if (is_numeric($entity)) {
    $entity = entity_revision_load($entity_type, $entity);
  }

  $entity_info = entity_get_info($entity_type);
  $id_field = $entity_info['entity keys']['id'];
  if ($message) {
    drupal_set_title(t('Revision of %title from %date', array(
      '%title' => $entity->title,
      '%date' => format_date($entity->timestamp),
    )), PASS_THROUGH);
  }

  // For markup consistency with other pages, use node_view_multiple() rather than node_view().
  $entities = entity_view($entity_type, array(
    $entity->$id_field => $entity,
  ), 'full');

  // Update the history table, stating that this user viewed this entity.
  // entity_tag_new($entity);

  return $entities;
}

/**
 * Access check for multiple permissions, return TRUE if at least one is TRUE.
 *
 * If return FALSE, the user will not have access to any of the LimeSurvey sync
 * administration menus ! It corresponds to the 'admin/structure/limesurvey'
 * menu root.
 *
 * @param $list
 *   An array of permissions to pass to the user_access() function. If empty,
 *   check all administration permissions on the LimeSurvey_sync module. Default
 *   is empty.
 * @param $account
 *   A user object. If NULL the connected user is used. Default is NULL.
 *
 * @return
 *   A Boolean. TRUE for access.
 */
function limesurvey_sync_api_multiple_access_check($list = array(), $account = NULL) {

  if ($list === array()) {
    // Load all main LimeSurvey module access.
    $list = array(
      'administer limesurvey sync survey',
      'administer limesurvey sync answer',
      'administer limesurvey sync token',
    );
  }
  foreach ($list as $right) {
    if (user_access($right, $account)) {
      return TRUE;
    }
  }
  return FALSE;
}

function limesurvey_sync_api_retreive_response_from_ref_entity($ref_entity) {
  if (empty($ref_entity->ls_sid)) {
    return array();
  }
  $response_entityname = limesurvey_sync_api_entityname('response');
  $ref_entityname = limesurvey_sync_api_entityname($ref_entity->ls_entity_type, $ref_entity->ls_sid);
  $ref_field_name = limesurvey_sync_api_et_fieldname($ref_entity->ls_entity_type, $ref_entity->ls_sid);
  $survey_typename = limesurvey_sync_api_typename('survey', $ref_entity->ls_sid);
  $ref_entity_info = entity_get_info($ref_entityname);
  $dp_key = $ref_entity_info['entity keys']['id'];
  $dp_key_value = $ref_entity->$dp_key;
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', $response_entityname)
  ->entityCondition('bundle', $survey_typename)
  ->fieldCondition($ref_field_name, 'target_id', $dp_key_value, '=' )
  //->range(0, 10)
  ->propertyOrderBy('dp_rid', 'DESC')
  ->addMetaData('account', user_load(1)); // Run the query as user 1.

  $result = $query->execute();
  $all_responses = array();
  if (isset($result[$response_entityname])) {
    $all_responses = entity_load($response_entityname, array_keys($result[$response_entityname]));
  }

  return $all_responses;
}

function limesurvey_sync_api_retreive_uid_from_ref_entity($ref_entity) {
  // @TODO try from the participant entity if token persistance enabled.
  $all_responses = limesurvey_sync_api_retreive_response_from_ref_entity($ref_entity);
  $all_uids = array();
  foreach ($all_responses as $dp_rid => $response_entity) {
    $all_uids[$dp_rid] = $response_entity->uid;
  }
  $unique_uids = array_unique($all_uids);
  if (count($unique_uids) == 1) {
    $uid = current($unique_uids);
  }
  elseif (empty($unique_uids)) {
    $uid = 0; // set as anonymous.
  }
  else {
    // Many uids associated to this reference, return an array.
    $uid = array_values($unique_uids);
  }
  return $uid;
}

/**
 * Return the LS survey ID from the entity type name.
 *
 * As an example, for 'limsurvey_sync_answer_123456' and 'answer', returns
 * '123456'.
 *
 * @param $entity_type
 *   The entity type of the entity.
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 *
 * @return
 *   A numeric, the LS survey ID or FALSE if error or NULL if invalid
 *   parametters.
 */
function limesurvey_sync_api_retrieve_ls_sid_from_entityname($entity_type, $ls_entity_type) {

  $with_suffix = array('token', 'answer');
  $without_suffix = array('survey', 'response');
  if (!empty($entity_type) && in_array($ls_entity_type, $with_suffix)) {
    $ls_sid = str_replace(limesurvey_sync_api_entityname($ls_entity_type, 'prefix'), '', $entity_type);
    if (empty($ls_sid) || !is_numeric($ls_sid)) {
      return FALSE;
    }
    return $ls_sid;
  }
  return NULL;
}

/**
 * Return the survey ID from the type name.
 *
 * Example : 'ls_survey_123456' will return '123456'.
 *
 * @param $typename
 *   An string. Corresponding to the variable name.
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 *
 * @return
 *   A numeric, the LS survey ID or FALSE if error or NULL if invalid
 *   parametters.
 */
function limesurvey_sync_api_retrieve_ls_sid_from_typename($typename, $ls_entity_type) {

  $entity_types = array('token', 'answer', 'survey');
  if (!empty($typename) && in_array($ls_entity_type, $entity_types)) {
    $ls_sid = str_replace(limesurvey_sync_api_typename($ls_entity_type, 'prefix'), '', $typename);
    if (empty($ls_sid) || !is_numeric($ls_sid)) {
      return FALSE;
    }
    return $ls_sid;
  }
  return NULL;
}

/**
 * Return the LS entity type from the entity type name.
 *
 * Example : 'limesurvey_sync_answer_123456' will return 'answer'.
 *
 * @param $entity_type
 *   The entity type of the entity.
 *
 * @return
 *   A string, the LS entity type or FALSE if error or NULL if invalid
 *   parametters.
 */
function limesurvey_sync_api_retrieve_ls_entity_type_from_entityname($entity_type) {

  if (empty($entity_type) || !is_string($entity_type)) {
    return FALSE;
  }
  $plit = explode('_', $entity_type);
  $count = count($plit);
  if ($count < 3 || strpos($entity_type, 'limesurvey_sync_') !== 0) {
    return FALSE;
  }
  $with_suffix = array('token', 'answer');
  $without_suffix = array('survey', 'response');
  $ls_entity_type = $plit[2];
  if (!in_array($ls_entity_type, array_merge($with_suffix, $without_suffix)) || (in_array($ls_entity_type, $without_suffix) && $count > 3) || (in_array($ls_entity_type, $with_suffix) && ($count > 4 || !is_numeric($plit[3])))) {
    return FALSE;
  }
  return $ls_entity_type;
}

/**
 * Return a dynamic title for answers and tokens.
 *
 * @param $entity
 *   The answer or token entity object.
 * @param $ls_key
 *   A string, The LS primary key fieldname on DP. It is provided into the
 *   entity_get_info, into 'entity keys' key as 'ls key'.
 *
 * @return
 *   A string, the tranlated title including the primary ls key value.
 */
function limesurvey_sync_api_default_label($entity, $ls_key) {

  $id = $entity->$ls_key;
  $arg = array(
    '@id' => $id,
    '@type' => drupal_ucfirst($entity->ls_entity_type),
    '@sid' => $entity->ls_sid,
  );
  return t('@type id #@id from the survey #@sid', $arg);
}

/**
 * List of the lss_status possible values.
 *
 * Used for compatibility with Views and Rules modules, throught
 * hook_entity_property_info().
 *
 * @return
 *   An array, the keys are the lss_status value and values are the human
 *   corresponding display message.
 */
function limesurvey_sync_api_lss_sync_status_list() {
  $status = array(
    LIMESURVEY_SYNC_SYNCHRONIZED,
    LIMESURVEY_SYNC_DESYNCHRONIZED,
    LIMESURVEY_SYNC_UNSYNCHRONIZED,
  );
  $list = array();
  module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_api.helpers');
  foreach ($status as $value) {
    $list[$value] = limesurvey_sync_api_display_field('survey', 'lss_sync', $value);
  }
  return $list;
}

/**
 * Return the object info for a LS entity type .
 *
 * From the LimesurveySyncInfos class, for building entty info or schema or root
 * uri.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 *
 * @return
 *   The entity info object.
 */
function limesurvey_sync_api_info_class_object($ls_entity_type = '') {
  if (empty($ls_entity_type)) {
    return FALSE;
  }
  $sub_entities = array('token', 'answer');
  $prefix = (!in_array($ls_entity_type, $sub_entities)) ? 'LimesurveySyncDpInfosLs' : 'LimesurveySyncDpInfosLsSubentity';
  $suffix = drupal_ucfirst(str_replace('_', '', $ls_entity_type));
  $class = $prefix . $suffix;
  $object = new $class();
  return $object;
}


/**
 * Create a new entity object for a LS entity type and a LS survey ID.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 * @param $ls_sid
 *   A numeric, the LS survey ID.
 *
 * @return
 *   An entity object.
 */
function limesurvey_sync_api_entity_class_object($ls_entity_type, $ls_sid = FALSE) {

  $class = 'LimesurveySync' . drupal_ucfirst(str_replace('_', '', $ls_entity_type));
  $values = array(
    'ls_sid' => $ls_sid,
  );
  $entity_type = limesurvey_sync_api_entityname($ls_entity_type, $ls_sid);
  // $object = (!empty($sid)) ? new $class(array('ls_sid' => $sid)) : new $class;
  $object = entity_create($entity_type, $values);
  return $object;
}

/**
 * Return a type name for a LS entity type and a LS survey ID.
 *
 * As an example : return 'ls_answer_123456' for 'answer' and '123456'.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 * @param $ls_sid
 *   A numeric, the LS survey ID or the string 'prefix' for returning the prefix
 *   ('ls_answer_' in the example). Or FALSE if an error occures.
 *
 * @return
 *   An entity object.
 */
function limesurvey_sync_api_typename($ls_entity_type, $ls_sid = FALSE) {

  if (empty($ls_sid)) {
    return FALSE;
  }
  $prefix = "ls_{$ls_entity_type}_";
  if ($ls_sid == 'prefix') {
    $ls_sid = '';
  }
  switch ($ls_entity_type) {
    case 'prefix':
    case 'survey':
    case 'answer':
    case 'token':
    case 'response':
      // Return 'ls_survey_123456' or 'ls_answer_123456', or 'ls_token_123456'.
      return "ls_{$ls_entity_type}_{$ls_sid}";
  }
  return FALSE;
}

/**
 * Return field name for a LS entity type and a LS survey ID.
 *
 * As an example : return 'field_ls_answer_123456' for 'answer' and '123456'. It
 * is used by the Field core module.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 * @param $ls_sid
 *   A numeric, the LS survey ID. Default is FALSE.
 *
 * @return
 *   Astring, the field name.
 */
function limesurvey_sync_api_et_fieldname($ls_entity_type, $ls_sid = FALSE) {
  return 'field_' . limesurvey_sync_api_typename($ls_entity_type, $ls_sid);
}

/**
 * Return entity name for a LS entity type and a LS survey ID.
 *
 * As an example : return 'limesurvey_sync_answer_123456' for 'answer' and
 * '123456'.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 * @param $ls_sid
 *   A numeric, the LS survey ID. Default is FALSE.
 *
 * @return
 *   Astring, the field name.
 */
function limesurvey_sync_api_entityname($ls_entity_type, $ls_sid = FALSE) {

  // If $ls_sid value is 'prefix : return the prefix entity name : as an example :
  // 'limesurvey_sync_answer_'.
  $with_suffix = array('token', 'answer');
  $without_suffix = array('survey', 'response');

  if (in_array($ls_entity_type, $with_suffix) && (empty($ls_sid) || (!is_numeric($ls_sid) && $ls_sid != 'prefix'))) {
    return FALSE;
  }
  if ($ls_sid == 'prefix') {
    $ls_sid = '';
  }
  $sid_suffix = (in_array($ls_entity_type, $with_suffix)) ? "_{$ls_sid}" : '';
  if (in_array($ls_entity_type, array_merge($with_suffix, $without_suffix))) {
    // Return 'limesurvey_sync_survey' or 'limesurvey_sync_answer_123456', or
    // or 'limesurvey_sync_token_123456'.
    return "limesurvey_sync_{$ls_entity_type}{$sid_suffix}";
  }
  return FALSE;
}

/**
 * Return database table name for a LS entity type and a LS survey ID.
 *
 * As an example : return 'limesurvey_sync_answer_123456_revision' for 'answer'
 * and '123456' and TRUE.
 *
 * @param $ls_entity_type
 *   A string, 'response', or 'answer', or 'token', or 'survey'.
 * @param $ls_sid
 *   A numeric, the LS survey ID. Default is FALSE.
 * @param $revision
 *   A boolean, if TRUE, return the revision table name. Default is FALSE.
 *
 * @return
 *   A string, the database table name.
 */
function limesurvey_sync_api_tablename($ls_entity_type, $ls_sid = FALSE, $revision = FALSE) {
  $entity_name = limesurvey_sync_api_entityname($ls_entity_type, $ls_sid);
  $revision_suffix = ($revision && $entity_name) ? '_revision' : '';
  return $entity_name . $revision_suffix;
}

/**
 * Return TRUE if the data is serialized.
 *
 * @param $str
 *   A string, the serialized data.
 *
 * @return
 *   A Boolean, TRUE if the data is serialized.
 */
function limesurvey_sync_api_valid_serialized_data($str) {
  $data = @unserialize($str);
  return ($str === 'b:0;' || $data !== false);
}
