<?php

/**
 * @file
 * Functions and classes for the survey entities.
 */

define('LIMESURVEY_SYNC_SURVEY_OPEN', 1);
define('LIMESURVEY_SYNC_SURVEY_SOON', 2);
define('LIMESURVEY_SYNC_SURVEY_EXPIRED', 3);
define('LIMESURVEY_SYNC_SURVEY_DESACTIVATED', 4);
define('LIMESURVEY_SYNC_SURVEY_UNKNOWN', 5);

/**
 * The entity info class for survey entities.
 */
class LimesurveySyncDpInfosLsSurvey extends LimesurveySyncDpInfosLs {
  public $ls_entity_type = 'survey';
  public $root_admin_ui = 'admin/structure/limesurvey/surveys';
  public $crud_path = 'admin/structure/limesurvey/surveys/manage';
  protected $revision = FALSE;
  protected $perm_view = TRUE;

  protected $bundle_of = 'limesurvey_sync_response';
  protected $bundle_wildcard = 'limesurvey_sync_survey';
  protected $menu_wildcard = '%limesurvey_sync_survey';
  protected $primary_key = 'dp_sid';
  protected $ls_key = 'ls_sid';
  protected $label = 'survey';
  protected $plural_label = 'surveys';
  protected $fieldable = FALSE;
  protected $status_col = FALSE;
  // If the comment are open or not.
  protected $comment_col = FALSE;
  // If the entity can be promoted to front page.
  protected $promote_col = FALSE;
  // If the entity can be set as sticky.
  protected $sticky_col = FALSE;
  // If the entity is fieldable.
  // colomns for bundles.
  protected $weight_col = TRUE;

  protected function schema_dp_table($revision = FALSE, $datas = array()) {
    $schema_table = array();
    $schema_table['description'] = 'Stores information about limeSurvey surveys.';
    $schema_table['unique'] = array(
      'dp_sid' => array(
        'dp_sid',
      ),
    );
    $schema_table['indexes'] = array(
      'type' => array(
        'type',
        'dp_sid',
      ),
    );

    $schema_table = array_merge(parent::schema_dp_table($revision), $schema_table);
    return $schema_table;
  }

  /**
   * Define schema for custom fields.
   *
   * each class should implement it.
   *
   * @return
   *   An array, the custom schema fields or an empty array if there is
   *   no fields.
   *
   * @see LimesurveySyncDpInfos::schema()
   */
  protected function schema_custom_fields($revision = FALSE, $datas = array()) {

    $schema = parent::schema_custom_fields($revision);
    $schema['ls_sid'] = array(
      'type' => 'int',
      'unsigned' => TRUE,
      'not NULL' => TRUE,
      'default' => 0,
      'description' => 'survey ID reference from the LimeSurvey site database',
    );
    $schema['lss_status'] = array(
      'type' => 'int',
      'unsigned' => TRUE,
      'not NULL' => TRUE,
      'default' => 1,
      'size' => 'tiny',
      'description' => 'Status of this survey (1 for open, 2 and 3 for closed, 4 for desactivated, 5 for error).',
    );

    return $schema;
  }

  protected function entity_property_info_custom_fields($datas = array()) {

    $info = parent::entity_property_info_custom_fields($datas);
    $info['ls_sid'] = array(
      'label' => t('DP LimeSurvey survey ID'),
      'description' => t('The LimeSurvey ID into the LimeSurvey software'),
      'type' => 'integer',
      'schema field' => 'ls_sid',
    );
    $info['lss_status'] = array(
      'label' => t('DP Survey status'),
      'description' => t('statut ID for the survey (1 is open, up is closed.'),
      'type' => 'list<int>',
      'options list' => 'limesurvey_sync_api_survey_lss_status_list',
      'schema field' => 'lss_status',
    );
    return $info;
  }

  /**
   * Provide default values on entity custom fields.
   *
   * Each class should implement it.
   *
   * @return
   *   An array, the defaut values keyed by there field name.
   */
  protected function entity_default_values_custom($datas = array()) {
    $default_values = parent::entity_default_values_custom();
    $default_values['ls_sid'] = 0;
    $default_values['lss_status'] = 3;
    return $default_values;
  }

}


/**
 * Class managing building survey entities.
 *
 * Building entity_info, entity_property_info, schema, entity default values and
 * updating entity table if needed. Take care of Drupal fields, LimeSurvey fixed
 * fields and Limesurver variables fields.
 */
class LimesurveySyncEntityLsSurvey extends LimesurveySyncEntityLs {
  public $type;
  public $label;
  public $ls_sid;
  public $ls_entity_type = 'survey';
}

/**
 * The Controller for survey entities
 */
class LimesurveySyncSurveyController extends LimesurveySyncController {
  public $ls_entity_type = 'survey';
  public $ls_key = 'ls_sid';

  /**
   * Provide the survey entity status.
   *
   * Status are 'open', 'soon', 'expired', 'desactivated', and 'unknown'.
   *
   * @param $entity
   *   The entity object.
   *
   * @return
   *   A numeric defined as a constant corresponding to a defined LS stage.
   */
  protected function lss_status($entity) {

    if (is_array($entity)) {
      $entity = (object) $entity;
    }
    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_survey');
    $survey_properties = limesurvey_sync_survey_properties($entity->ls_sid);
    if ($survey_properties) {
      if (!$survey_properties['active']) {
        return LIMESURVEY_SYNC_SURVEY_DESACTIVATED;
      }
      elseif ((empty($survey_properties['startdate']) || $survey_properties['startdate'] <= REQUEST_TIME) && (empty($survey_properties['expires']) || $survey_properties['expires'] > REQUEST_TIME)) {
        return LIMESURVEY_SYNC_SURVEY_OPEN;
      }
      elseif ($survey_properties['startdate'] > REQUEST_TIME) {
        return LIMESURVEY_SYNC_SURVEY_SOON;
      }
      elseif ($survey_properties['expires'] < REQUEST_TIME) {
        return LIMESURVEY_SYNC_SURVEY_EXPIRED;
      }
    }
    return LIMESURVEY_SYNC_SURVEY_UNKNOWN;
  }

  /**
   * Builds a structured array representing the entity's content.
   *
   * The content built for the entity will vary depending on the $view_mode
   * parameter.
   *
   * @param $entity
   *   An entity object.
   * @param $view_mode
   *   View mode, e.g. 'full', 'teaser'...
   * @param $langcode
   *   (optional) A language code to use for rendering. Defaults to the global
   *   content language of the current request.
   *
   * @return
   *   The renderable array.
   */
  public function buildContent($entity, $view_mode = 'full', $langcode = NULL, $content = array()) {

    // Our additions to the $build render array
    $wrapper = entity_metadata_wrapper('limesurvey_sync_survey', $entity);

    $content['ls_sid'] = array(
      '#theme' => 'field',
      '#weight' => 0,
      '#title' => t('Survey ID'),
      '#access' => TRUE,
      '#label_display' => 'inline',
      '#view_mode' => 'full',
      '#language' => LANGUAGE_NONE,
      '#field_name' => 'ls_sid',
      '#field_type' => 'text',
      '#entity_type' => 'limesurvey_sync_survey',
      '#bundle' => 'limesurvey_sync_survey',
      '#items' => array(
        array(
          'value' => $entity->ls_sid,
        ),
      ),
      '#formatter' => 'text_default',
      0 => array(
        '#markup' => check_plain($entity->ls_sid),
      ),
    );

    $introduction = '<p>' . t('All functionalities may not be availables from this drupal site, depending on specific survey settings from your LimeSurvey admin interface.') . '<br /><strong>' . t(
        'The table below shows which features are availables and how making then availables, setting each LimeSurvey parameter using the [edit] link into the right colomn.') . '</strong></p>';

    $content['introduction'] = array(
      // '#theme' => 'markup',
      '#markup' => $introduction,
      '#weight' => -4,
    );

    $form_state = array();
    $form_state['build_info']['args'] = array($entity, TRUE);
    $form_state['values'] = array();
    $form_state += form_state_defaults();
    module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_survey');

    $element = array(
      '#type' => 'form',
      '#tree' => FALSE,
      '#id' => 'limesurvey_sync_survey_sid_form',
      'content_survey' => limesurvey_sync_survey_sid_form(array(), $form_state, $entity, FALSE),
      '#weight' => -10,
    );
    form_builder('limesurvey_sync_survey_sid_form', $element, $form_state);
    $content['limesurvey_survey_infos'] = $element;

    $table_survey = limesurvey_sync_survey_view_properties($entity);
    $content['survey_properties'] = $table_survey;

    $response_entityname = limesurvey_sync_api_entityname('response', $entity->ls_sid);
    $response_entity = entity_create($response_entityname, array('ls_sid' => $entity->ls_sid, 'type' => $entity->type));
    if (entity_access('create', $response_entityname, $response_entity)) {
      // Prepare link 'To response to the survey'.
      $new_windows = array('attributes' => array('onclick' => 'window.open(this.href); return false;'));
      $entity_info = entity_get_info('limesurvey_sync_response');
      $add_response_uri = $entity_info['crud path'] . '/add/' . str_replace('_', '-', $entity->type);
      $content['new_answer'] = array(
        '#markup' => l(t('To response to the survey'), $add_response_uri, $new_windows),
        '#weight' => 3,
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      );
    }

    if (user_access('administer limesurvey sync survey')) {
      // Prepare links to graphics answers on the LimeSurey website.
      // Load file for the limesurvey_sync_link_to_ls() function.
      module_load_include('inc', 'limesurvey_sync', 'limesurvey_sync');
      $stat_link = limesurvey_sync_link_to_ls(t('Responses on real time graphic statistics'), $entity->ls_sid, 'statistics');
      $content['ls_graph'] = array(
        '#markup' => $stat_link,
        '#weight' => 4,
        '#prefix' => '<p>',
        '#suffix' => '</p>',
      );
    }

    $build = parent::buildContent($entity, $view_mode, $langcode, $content);

    return $build;
  }

  /**
   * Save.
   *
   * On saving a survey : create or update the entity_info(), variables, and
   * database tables and fields for the answer (and token if needed).
   */
  public function save($entity, DatabaseTransaction $transaction = NULL) {

    // Set a type.
    $entity->type = limesurvey_sync_api_typename($entity->ls_entity_type, $entity->ls_sid);

    parent::save($entity, $transaction);

    drupal_get_schema(NULL, TRUE);
    entity_info_cache_clear();
    variable_set('menu_rebuild_needed', TRUE);

    // Add responses fields.
    $response = limesurvey_sync_api_info_class_object('response');
    $response->update_table();

    // Set default options.
    $variable_name = "limesurvey_sync_survey_{$entity->ls_sid}_options";
    $ls_options = variable_get($variable_name, array());
    if (empty($ls_options)) {
      module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_invitation.options');
      $new_settings = limesurvey_sync_invitation_options_default($entity->ls_sid, $ls_options);
      variable_set($variable_name, $new_settings);
    }
  }

  /**
   * Delete.
   *
   * On deleting a survey : Delete the variables, fields and database tables for
   * the answer (and token if needed).
   */
  public function delete($ids, DatabaseTransaction $transaction = NULL) {
    $entities = $ids ? $this->load($ids) : FALSE;
    if ($entities) {
      // Delete associated answer and token entities and fields.
      foreach ($entities as $id => $entity) {
        // Select the response entities associated to this survey.
        db_delete(limesurvey_sync_api_tablename('response'))->condition('type', $entity->type)->execute();
        field_attach_delete_bundle($this->entityType, $entity->type);

        foreach (array('answer', 'token') as $ls_type) {
          $ref_tablename = limesurvey_sync_api_tablename($ls_type, $entity->ls_sid);
          if (db_table_exists($ref_tablename)) {
            db_drop_table($ref_tablename);
          }
          // Delete revision table to.
          $ref_tablename_revision = limesurvey_sync_api_tablename($ls_type, $entity->ls_sid, TRUE);
          if (db_table_exists($ref_tablename_revision)) {
            db_drop_table($ref_tablename_revision);
          }
          // Delete associated variables.
          $ref_entityname = limesurvey_sync_api_entityname($ls_type, $entity->ls_sid);
          variable_del("{$ref_entityname}_options");
          variable_del("{$ref_entityname}_ls_fields");
        }
      }
      parent::delete($ids, $transaction);
    }
    field_purge_batch(100);
  }

}

/**
 * The UI Controller for survey entities
 */
class LimesurveySyncSurveyUIController extends LimesurveySyncUIController {
  public $ls_entity_type = 'survey';
  // Survey are added by a user.
  protected $has_view_page = TRUE;
}

/**
 * Menu argument loader; Load a survey from menu wildcard.
 *
 * @param $name
 *   The machine-readable name of a survey to load.
 *
 * @return A survey object or FALSE if $name does not exist.
 */
function limesurvey_sync_survey_load($name) {
  return limesurvey_sync_survey_get_surveys(strtr($name, array('-' => '_')));
}


/**
 * Gets an array of all imported surveys, keyed by the survey ID.
 *
 * @param $ls_sid
 *   A numeric, the LS survey  ID, or 'all' for all surveys. Default is 'all'.
 * @return
 *   LimeSurveySyncSurvey[] Depending whether $type isset, an array of surveys ID.
 */
function limesurvey_sync_survey_get_survey_by_sid($ls_sid = 'all') {
  // entity_load will get the Entity controller for our survey entity and call the load
  // function of that object - we are loading entities by name here.
  $surveys = entity_load_multiple_by_name('limesurvey_sync_survey', FALSE);
  $return = array();
  foreach ($surveys as $type => $survey) {
    $return[$survey->ls_sid] = $survey;
  }
  return ($ls_sid == 'all') ? $return : ((empty($ls_sid) || !is_numeric($ls_sid) || !isset($return[$ls_sid])) ? FALSE : $return[$ls_sid]);
}

function limesurvey_sync_api_survey_tab($entity_type, $survey_type) {

  if (!is_object($survey_type)) {
    $survey = limesurvey_sync_survey_get_surveys($survey_type);
  }
  else {
    $survey = $survey_type;
  }
  if (entity_access('view', $entity_type, $survey)) {
    return entity_view($entity_type, array(
      $survey,
    ));
  }
  else {
    return array(
      '#markup' => t('Administration for the survey @label', array(
        '@label' => $survey->label,
      )),
    );
  }
}

/**
 * Generates the LimeSurvey survey editing form.
 */
function limesurvey_sync_survey_form($form, &$form_state, $survey, $op = 'edit') {

  //$form['#limesurvey_sync_survey'] = $survey;

  // Load form for selecting ls_sid.
  module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_survey');
  $form = array_merge($form, limesurvey_sync_survey_sid_form($form, $form_state, $survey));

  $form['ls_sid'] = array(
    '#type' => 'hidden',
    '#value' => $survey->ls_sid,
  );

  // Add options.
  module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_invitation.options');
  $form += limesurvey_sync_invitation_options_form(array(), $form_state, $survey->ls_sid);
  //$form['#submit'][] = 'limesurvey_sync_survey_form_submit';
  $form['#submit'][] = 'limesurvey_sync_invitation_options_form_submit';

  $form['label'] = array(
    '#title' => t('Label'),
    '#type' => 'textfield',
    '#default_value' => $survey->label,
    '#description' => t('The human-readable name of this LimeSurvey survey.'),
    '#required' => TRUE,
    '#size' => 30,
    '#weight' => -10,
  );
  $form['weight'] = array(
    '#type' => 'weight',
    '#title' => t('Weight'),
    '#default_value' => (isset($survey->weight)) ? $survey->weight : 0,
    '#description' => t('When showing surveys, those with lighter (smaller) weights get listed before profiles with heavier (larger) weights.'),
    '#weight' => 10,
  );

  if (count(limesurvey_sync_survey_get_surveys()) > 1) {
    // Show only the weight when there is at least 2 surveys.
    $form['weight']['#access'] = FALSE;
  }

  $form['description'] = array(
    '#title' => t('Description'),
    '#type' => 'textarea',
    '#default_value' => $survey->description,
    '#description' => t('Describe this survey. The text will be displayed on the <em>Add new response</em> page.'),
  );

  $form['submission']['help'] = array(
    '#type' => 'textarea',
    '#title' => t('Explanation or submission guidelines'),
    '#default_value' => (empty($survey->is_new)) ? $survey->help : t('Submit this page in order to access to the survey form.'),
    '#description' => t('This text will be displayed at the top of the page when creating or editing response of this survey.'),
  );

  return $form;
}

/**
 * Form API validate callback for the limesurvey_sync_survey form.
 */
function limesurvey_sync_survey_form_validate($form, &$form_state) {
  $sid = $form_state['values']['ls_sid'];
  $target = 'surveys][ls_sid';
  if (empty($sid) || !is_numeric($sid)) {
    form_set_error('surveys', t('The Survey ID is not valid.'));
  }
  else {
    $unchanged_survey = $form_state['limesurvey_sync_survey']; //$form['#limesurvey_sync_survey'];
    $is_new = (!empty($unchanged_survey->is_new));
    module_load_include('inc', 'limesurvey_sync', 'limesurvey_sync');
    module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_survey');
    $this_survey = limesurvey_sync_survey_properties($sid);
    if (!$this_survey) {
      $message = t('The survey id %sid is not present into your LimeSurvey site', array(
        // @TODO : add LimeSurvey site link.
        '%sid' => $sid,
      ));
      limesurvey_sync_survey_error_or_message($is_new, $target, $message);
    }
    else {
      $this_survey = limesurvey_sync_survey_properties($sid);
      $existing_survey = limesurvey_sync_survey_get_survey_by_sid($sid);
      // A survey entity should be a unique sid.
      // Check if this sid does already exist.
      if ($existing_survey && ($is_new || $existing_survey->type != $unchanged_survey->type)) {
        // If we create a new survey entity, prevent for dooblon.
        $doubloon_link = (entity_access('view', 'limesurvey_sync_survey', $existing_survey)) ? t('the !entity survey', array(
          '!entity' => l($existing_survey->label, 'admin/structure/limesurvey/view/' . $existing_survey->type),
        )) : t('an unauthorized access survey');
        form_set_error($target, t('This survey ID is already taken by !survey. An alternative way is !copy from the LimeSurvey user interface.', array(
          '!survey' => $doubloon_link,
          '!copy' => limesurvey_sync_link_to_ls(t('to copy this survey'), NULL, 'newsurvey', 'copy'),
        )));
      }
      else {
        // Test for compatibility, active and tokens table.
        // Call error for new entity and display only message for existing
        // entity, it will allow entity edition on archive survey entitys.
        if (!$this_survey['active']) {
          $message = t('This survey is not currently active, you can !activate on the LimeSurvey user interface.', array(
            '!activate' => limesurvey_sync_link_to_ls(t('activate it'), $sid, 'activate'),
          ));
          limesurvey_sync_survey_error_or_message($is_new, $target, $message);
        }

        if (!$is_new && $sid != $unchanged_survey->ls_sid) {
          form_set_error($target, t('You can not change the survey ID. Create a <a href="@url">new LimeSurvey survey</a> instead.', array(
            '@url' => url('admin/structure/limesurvey/add'),
          )));
        }
      }
    }
  }
}

function limesurvey_sync_api_survey_lss_status_list() {
  $status = array(
    LIMESURVEY_SYNC_SURVEY_OPEN,
    LIMESURVEY_SYNC_SURVEY_SOON,
    LIMESURVEY_SYNC_SURVEY_EXPIRED,
    LIMESURVEY_SYNC_SURVEY_DESACTIVATED,
    LIMESURVEY_SYNC_SURVEY_UNKNOWN,
  );
  $list = array();
  module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_api.helpers');
  foreach ($status as $value) {
    $list[$value] = limesurvey_sync_api_display_field('survey', 'lss_status', $value);
  }
  return $list;
}

function limesurvey_sync_survey_ls_language_list($ls_sid = NULL) {
  module_load_include('inc', 'limesurvey_sync', 'ls_api/limesurvey_sync_ls_api_survey');
  // $ls_sid can be 'view' or 'edit'.
  if (empty($ls_sid) || !is_numeric($ls_sid)) {
    $surveys = limesurvey_sync_survey_properties('all');
  }
  else {
    $one_survey = limesurvey_sync_survey_properties($ls_sid);
    $surveys = array($ls_sid => $one_survey);
  }
  // languages for all surveys.
  $all_ls_lang = array();
  foreach ($surveys as $sid => $datas) {
    $all_ls_lang = array_merge($all_ls_lang, array_keys($datas['#lang']));
  }

  $ls_langs = array_flip($all_ls_lang);
  $lang_list = array();
  module_load_include('inc', 'limesurvey_sync', 'api/limesurvey_sync_api.helpers');
  foreach ($ls_langs as $ls_lang_code => $index) {
    $lang_list[$ls_lang_code] = limesurvey_sync_api_display_field('limesurvey_sync_survey', 'ls_lang', $ls_lang_code, 'value', $ls_sid);
  }

  return $lang_list;
}

/**
 * Return a drupal LS field for display.
 *
 * As an example, return timestamp as formated date, LS language code as human
 * language, etc...
 *
 * @param $entity_type
 *   A string. The entity type (ex : 'limesurvey_sync_answer_xxxxx').
 * @param $field
 *   A string. The Drupal ls_field to display (ex : 'startlanguage').
 * @param $value
 *   A string or any. corresponding Drupal LS field to display (ex:
 *   'de-informal').
 *
 * @return
 *   A string. The value ready to be displayed.
 */
function limesurvey_sync_survey_display_field($field, $value, $return = 'value', $ls_sid = FALSE) {

  if ($field == 'lss_status') {
    switch ($value) {
      case LIMESURVEY_SYNC_SURVEY_OPEN:
        switch ($return) {
          case 'help':
            return t('The survey is open, users can add responses.');
          case 'css':
            return 'limesurvey-sync-field-lss-status-survey-open';
          default:
            return t('Open');
        }

      case LIMESURVEY_SYNC_SURVEY_SOON:
        switch ($return) {
          case 'help':
            return t('The survey will be open in the future.');
          case 'css':
            return 'limesurvey-sync-field-lss-status-survey-soon';
          default:
            return t('Not open yet');
        }

      case LIMESURVEY_SYNC_SURVEY_EXPIRED:
        switch ($return) {
          case 'help':
            return t("The survey has expired, users can't add responses.");
          case 'css':
            return 'limesurvey-sync-field-lss-status-survey-expired';
          default:
            return t('Expired');
        }

      case LIMESURVEY_SYNC_SURVEY_DESACTIVATED:
        switch ($return) {
          case 'help':
            return t('The survey is desactivated.');
          case 'css':
            return 'limesurvey-sync-field-lss-status-survey-desactivated';
          default:
            return t('Closed');
        }

      case LIMESURVEY_SYNC_SURVEY_UNKNOWN:
        switch ($return) {
          case 'help':
            return t('Unknown survey status.');
          case 'css':
            return 'limesurvey-sync-field-lss-status-survey-unknown';
          default:
            return t('Unknown');
        }
    }
  }

  return $value;
}
